"""
GenomeLibrary is a more complex version of GenomeList:
 - support for genome sequences
 - genes are located inside the sequence
"""


from __future__ import annotations
from typing import Optional, List, NamedTuple

from numpy.random import Generator
from Bio.Seq import Seq
from Bio.Align import PairwiseAligner # Bio.Blast.Record.Alignment

from ...host import Host
from ...module import Module
from ...events import EventEmitter, EventLogger
from ...utils import alldef, coalesce, first
from ..records.gene.gene import Gene
from ..records.gene.localized_gene import LocalizedGene
from ..all_events import InsertGeneEvent, RemoveGeneEvent, AlterGenePromoterEvent



class PrimerMatch (NamedTuple) :
    loc_start: int # Match goes from start (inclusive) to end (exclusive).
    loc_end: int
    success: float # (0,1)



class GenomeLibrary (Module) :
    """
    GenomeLibrary is a module that models the genome in a simplified way to ease interpretation.
    It stores a sequence of base pairs and provides a Gene interface that reads from the sequence
    at the gene's locus.

    The Sequence is stored as a base-pair list, whereas genes are stored as locations inside
    that sequence.
    """

    # bg_size: int # Amount of background gene bases.

    # bg_gc_content: float # Ratio of GC-content in the background genome. Range: [0,1]

    sequence: Seq

    locgenes: List[LocalizedGene]



    def make ( self,
        sequence:Optional[Seq] = None,
        bg_size:Optional[int] = None,
        bg_gc_content:Optional[float] = None,
        bg_rnd:Optional[Generator] = None,
        locgenes:List[LocalizedGene] = []
    ) -> None :

        # Init 1: pre-defined sequence
        if alldef( sequence ) :
            self.sequence = sequence

        # Init 2: new sequence
        elif alldef( bg_rnd, bg_size, bg_gc_content ) :
            self.sequence = make_background_seq( size=bg_size, gc_content=bg_gc_content, rnd=bg_rnd )

        # Failed Init
        else :
            raise Exception ("Creating GenomeLibrary module without either pre-defined or generated sequence.")

        # Add the locgenes and check if they are matching.
        self.locgenes = []
        for lg in locgenes :
            new_loc_gene = LocalizedGene(
                name=lg.name, seq=self.sequence,
                start_loc=lg._start_loc, prom_len=lg.prom_len, orf_len=lg.orf_len
            )
            self.locgenes.append( new_loc_gene )



    def copy ( self, ref:'GenomeLibrary' ) -> None :
        self.sequence = ref.sequence
        self.locgenes = ref.locgenes.copy()



    def bind ( self, host:Host ) -> None :
        host.observe( InsertGeneEvent, self.listen_insert_gene )
        host.observe( RemoveGeneEvent, self.listen_remove_gene )
        host.observe( AlterGenePromoterEvent, self.listen_alter_gene_promoter )



    def sync ( self, emit:EventEmitter, log:EventLogger ) -> None :
        pass # Nothing to sync.



    @property
    def genes ( self ) -> List[Gene] :
        return self.locgenes



    def listen_insert_gene ( self, event:InsertGeneEvent, emit:EventEmitter, log:EventLogger ) -> None :

        # If no locus is specified, then append it to the sequence.
        loc:int = coalesce( event.locus, len(self.sequence) )

        self.insert_gene( gene=event.gene, loc=loc )
        log( "GenomeLibrary: added gene={} loc={}".format(
            event.gene.name, event.locus
        ))



    def listen_remove_gene ( self, event:RemoveGeneEvent, emit:EventEmitter, log:EventLogger ) -> None :
        """ Remove a gene from the library and sequence. """
        del self.genes[event.gene]
        log( "GenomeLibrary: removed gene={}".format(event.gene.name) )



    def listen_alter_gene_promoter ( self, event:AlterGenePromoterEvent, emit:EventEmitter, log:EventLogger ) -> None :
        """
        When a gene promoter is altered, we replace the associated part in the sequence.
        TODO: No knock-outs are being made on this operation.
        """
        found_gene = first( self.locgenes, lambda lg : lg == event.gene )
        if found_gene is not None :
            self.alter_sequence( found_gene.start_loc, found_gene.prom_len, event.new_promoter )
            log( "GenomeLibrary: Changed promoter of gene={} in sequence.".format(event.gene.name) )



    def get_genes_by_name ( self, name:str ) -> List[Gene] :
        found_genes = [ gene for gene in self.locgenes if name == gene.name ]
        return found_genes



    def insert_gene ( self, gene:Gene, loc:int ) :
        """
        Insert a gene and its sequence at a specific location in the sequence.
        """
        # Insert the sequence at the insertion point.
        self.sequence = self.sequence[0:loc] + gene.prom + gene.orf + self.sequence[loc:None]

        # Remove genes that start before the insertion and end afterwards (knockout)
        is_knocked_out = lambda lg : lg.start_loc < loc and lg.end_loc > loc
        self.locgenes[:] = [ lg for lg in self.locgenes if not is_knocked_out(lg) ]
            # Replace the list of locgenes with the filtered locgenes, while keeping the list reference.

        # Shift the location of all genes after the insertion point.
        new_gene_len = len(gene.prom) + len(gene.orf)
        for lg in self.locgenes :
            lg.seq = self.sequence # Update to use this new sequence (since they are immutable)
            if lg.start_loc >= loc :
                lg.start_loc = lg.start_loc + new_gene_len

        # Annotate the new sequence as a localized gene.
        new_loc_gene = LocalizedGene(
            name=gene.name, seq=self.sequence,
            start_loc=loc, prom_len=len(gene.prom), orf_len=len(gene.orf)
        )
        self.locgenes.append( new_loc_gene )



    def alter_sequence ( self, old_loc:int, old_len:int, new_seq:Seq ) :
        self.sequence = self.sequence[0:old_loc] + new_seq + self.sequence[old_loc+old_len:None]



    def calc_primer_matches ( self, primer:Seq ) -> List[PrimerMatch] :
        """
        Return the insertion sites a primer can have, alongside with success rate.
        The insertion site is located right after a primer match.
        """
        matches: List[PrimerMatch] = []
        targets: List[Seq] = [ self.sequence ] # List of sequences to check.
        for target in targets :
            if len(target) > 0 :
                match = find_best_primer_match( target, primer )
                if match is not None :
                    matches.append( match )
                    targets.append( target[None:match.loc_start] ) # Also search left of match.
                    targets.append( target[match.loc_end:None] ) # Also search right of match.
        return matches



def make_background_seq ( size:int, gc_content:float, rnd:Generator ) -> Seq :
    gc = gc_content
    at = 1 - gc_content
    seq_array = rnd.pick_choices( choices=["A","C","G","T"], amount=size, weights=[at/2,gc/2,gc/2,at/2] ) # 'ATCG' is ArrayLike
    return Seq( "".join(seq_array) )



def check_primer_integrity ( primer:Seq ) -> bool :
    """ Non-deterministic check if the primer itself is well built. """
    return True



def find_best_primer_match ( sequence:Seq, primer:Seq ) -> Optional[PrimerMatch] :
    """ Find best match for a template sequence.
    TODO: Primer matching is dependent on host. algo_params:Optional[Dict] = None
    """

    # The primer matches with the complement sequence.
    comp_seq = sequence.complement()

    # Build the aligner for matching.
    aligner = PairwiseAligner()
    #aligner = extend( aligner, algo_params )
    aligner.mode = 'global' # target_end_gap_score, query_end_gap_score
    aligner.match_score = 1
    aligner.mismatch_score = -1
    aligner.gap_score = -1 # open_gap_score, extend_gap_score
    aligner.query_end_gap_score = 0

    # Get the alignment with the best score, anywhere on the sequence.
    all_alignments = aligner.align( comp_seq, primer )
    best_alignment = next(iter(sorted(all_alignments)))
    #print("score:" + str(best_alignment.score) + "\n" + str(best_alignment))
    if best_alignment is not None and best_alignment.score > 0 :
        return PrimerMatch(
            loc_start= best_alignment.aligned[0][0][0], # start of first chunk in target
            loc_end= best_alignment.aligned[0][-1][1], # end of last chunk in target
            success= best_alignment.score / len(primer)
        )
    return None



# TODO: Unused method that would create a full sequence by taking a list of genes and filling the
#   spaces in between with background bases. Since the background sequence is created at the
#   beginning I am not using this method.
# def build_sequence_from_gene_list ( self ) -> Seq :
#     rnd = self.host.make_generator()
#     gc = self.bg_gc_content
#     at = 1 - self.bg_gc_content
#
#     # Write the sequence from begin to end. Genes have their sequence copied and random bases
#     # are used to fill non-coding positions. Overlapping genes have undefined behaviour.
#     # TODO: When gene positions overlap it actually writes both gene sequences in sequence but
#     # in future this behaviour should be improved. This deals with validity of genes in
#     # overlapping regions.
#     locgenes = self.genes.values()
#     locgenes.sort( key=lambda el: el.loc )
#
#     seq = Seq()
#     cur = 0
#     bg_size = 0 # Current amount of background bases.
#     for locgene in locgenes :
#
#         # If the gene starts later, then fill the space with random (background) bases.
#         filler = locgene.loc - cur
#         if filler > 0 :
#             new_bases = rnd.choice( 'ATCG', size=filler, p=[at,at,gc,gc] )
#             bg_size += filler
#             seq += "".join(new_bases)
#
#         # Copy the gene sequence.
#         seq = seq + locgene.gene.get_prom() + locgene.gene.get_orf()
#
#     # At the end, if the minimum of background bases was not filled, to that now.
#     filler = min( 0, self.bg_min_size - bg_size )
#     if filler > 0 :
#         new_bases = rnd.choice( 'ATCG', size=filler, p=[at,at,gc,gc] )
#         seq += "".join(new_bases)
#
#     return seq



"""
Registries provide an efficient way to store Records that are used by multiple Hosts.
That way, each host only needs to store a reference to each registered object (for example
a gene) instead to storing and copying the entire gene.
When editing an object in the registry, the editor should make sure to practice copy-on-write.

Entries of a registry must be copy-able.
"""

from typing import Set

from .record import Record



class Registry :
    """ A Registry holds multiple Entries. """

    records : Set[Record]

    def __init__ ( self ) :
        self.records = set()


    def insert ( self, rec: Record ) -> None :
        """ Insert a new record into the Registry. """
        self.records.add(rec)


    def clone ( self, rec: Record ) -> Record :
        """ Clone a Record inside the Registry and return it. """
        if rec in self.records :
            cloned_rec = rec.clone()
            self.insert(cloned_rec)
            return cloned_rec
        raise Exception("Tried to clone a non-registered record.")






from typing import Union, Generic, TypeVar, List
from dataclasses import dataclass
import pathlib
import numpy as np
import pandas as pd
from openpyxl import load_workbook

from pandas import DataFrame, Series
import matplotlib.pyplot as plt



T = TypeVar('Value')



class SimulationException (Exception) :
    """ Base class for all simulation exceptions. """
    pass



@dataclass(frozen=True)
class Outcome (Generic[T]) :
    """ Return from operations with any type of result. """
    value: T
    error: Union[str,None] = None

    def has_error ( self ) -> bool :
        return self.error is not None

    def succeeded ( self ) -> bool :
        return not self.has_error()



@dataclass(frozen=True)
class DataOutcome (Outcome) :
    """
    Holds the dataframe of a simulation. Has methods to access whether it worked successfully, to
    print the data or to store it in files.
    """
    value: Union[DataFrame,Series]

    def __str__ ( self ) -> str :
        return str(self.value)

    # def has_error ( self ) -> bool :
    #     return self.error is not None

    # def succeeded ( self ) -> bool :
    #     return not self.has_error()

    def display_data ( self ) -> None :
        if self.value is not None :
            # old_max_rows = pd.options.display.max_rows
            # old_max_columns = pd.options.display.max_columns
            # pd.options.display.max_rows = 100
            # pd.options.display.max_columns = 30
            print(self.value)
            # pd.options.display.max_rows = old_max_rows
            # pd.options.display.max_columns = old_max_columns
        else :
            print("Outcome is empty.")

    def export_data ( self, filepath ) -> None :
        abs_filepath = pathlib.Path(filepath).resolve()
        abs_filepath.parent.mkdir(parents=True, exist_ok=True) # Build non-existing parent dirs.
        self.value.to_csv( abs_filepath, index=False )
        print("Data exported to: {}".format(abs_filepath))

    def append_data2xlsx ( self, filepath , sheet) -> None :
        ''' Append data to existing excel file, if it does not exist create a new one.
            If the file exists, it will also copy metadata from the 'Reference' sheet to the new sheet.
            
            filepath: path to the excel file
            sheet: name of the sheet to append the data to
        '''
        abs_filepath = pathlib.Path(filepath).resolve()

        # rename columns of dataframe myDat
        self.value.rename(columns={'t': 'time', 'S': 'Substrate'}, inplace=True)

        if not abs_filepath.exists():
            # Create a new Excel file with the DataFrame
            with pd.ExcelWriter(abs_filepath, engine='openpyxl', mode='w') as writer:
                self.value.rename(columns={'t': 'time', 'S': 'Substrate'}, inplace=True)
                self.value.to_excel(writer, sheet_name=sheet, index=False, startrow=0, startcol=0)
                return
        else:
            # Excel file for results already exists
        # Load the existing workbook
            with pd.ExcelWriter(abs_filepath, engine='openpyxl', mode='a', if_sheet_exists='overlay') as writer:
                # check if 'Reference' sheet exists
                if 'Reference' in writer.book.sheetnames:
                    # load metadata from 'Reference' sheet
                    ref_sheet = writer.book['Reference']
                    data_range = ref_sheet['A1:C15']  # Example range, adjust as needed
                    # Convert the extracted range into a list of lists
                    data_list = [[cell.value for cell in row] for row in data_range]
                    # Convert the list of lists into a DataFrame
                    meta_df = pd.DataFrame(data_list, columns=['Metadata: Key', 'Metadata: Value', 'Metadata: Comment'])  # Adjust column names as needed
                    meta_df.to_excel(writer, sheet_name=sheet, index=False, header=False)
                    # Append the new data to
                    self.value.to_excel(writer, sheet_name=sheet, index=False, startrow=0, startcol=4)
                else:
                    self.value.to_excel(writer, sheet_name=sheet, index=False, startrow=0, startcol=4)                
                return

        # Save the changes
        writer.close()
        # # print(f"Data appended to: {abs_filepath} in sheet {sheet}")



@dataclass(frozen=True)
class DataWithPlotOutcome (DataOutcome) :

    def make_plot ( self ) -> plt.Figure :
        """
        Plotting with pyplot is unfortunately unintuitive. You cannot display a single figure by
        using the object-oriented API. When you do `plt.subplots` (or create a plot by any other
        means) it will be stored in the global context. You can only display things from the glbbal
        context, and displaying it will remove it from there.
        """
        fig, ax = plt.subplots()
        self.value.plot( y=self.value.columns, use_index=True, subplots=True, ax=ax )
        return fig

    def display_plot ( self ) -> None :
        plt.rcdefaults()
        self.make_plot()
        plt.show()

    def export_plot ( self, filepath ) -> None :
        abs_filepath = pathlib.Path(filepath).resolve()
        abs_filepath.parent.mkdir(parents=True, exist_ok=True) # Build non-existing parent dirs.
        plt.rcdefaults()
        fig = self.make_plot()
        plt.savefig(abs_filepath)
        plt.close(fig)
        print("Plot exported to: {}".format(abs_filepath))



def combine_data ( outcomes:List[DataOutcome] ) -> DataOutcome :
    """
    Combine multiple DataOutcome into one big table. Can combine multiple Series and Dataframes.
    TODO: Should it extract the error of each outcome? Right now it is being ignored.
    """

    # First convert all outcomes into dataframes (convert Series and pass DataFrames)
    dfs = []
    for outcome in outcomes :
        if isinstance( outcome.value, pd.DataFrame ) :
            dfs.append( outcome.value )
        elif isinstance( outcome.value, pd.Series ) :
            dfs.append( pd.DataFrame([outcome.value]) )

    # Make the join and outcome a dataframe.
    cout = pd.concat(dfs)
    cout.reset_index( drop=True, inplace=True )
    return DataOutcome( value=cout )


def add_dfcolumn(Vector1: np.ndarray, Array: np.ndarray, RowIndex: np.ndarray, n_rows: int) -> pd.Series:
    """
    Build a pandas Series of length n_rows (optionally using provided index) with object dtype.
    At positions given by RowIndex (list/array of index labels or integer positions)
    the Series contains dictionaries {'time': Vector1, 'signal': signal_array}.
    All other entries are NaN.

    Vector1 : 1D numpy array (e.g., time axis)
    Array : 2D array-like with shape (num(RowIndex), len(Vector1))
    RowIndex : row indices (labels) or integer positions
    n_rows : int, length of the output Series
    """
    #   test whether Array matches RowIndex length
    Array = np.asarray(Array)
    if Array.shape[0] != len(RowIndex):
        raise ValueError("Array shape[0] must match length of RowIndex")

    # create object-typed Series with given index (or 0..n_rows-1)
    s = pd.Series([np.nan] * n_rows, index=range(n_rows), dtype=object)

    # assign dictionaries at requested locations
    for loc, sig in zip(RowIndex, Array):
        # allow loc to be an index label (in idx) or an integer position
        s.at[loc] = {'time': np.asarray(Vector1), 'signal': np.asarray(sig)}

    return s


"""
An experiment that uses hosts which implements all currently existing modules.
"""

from __future__ import annotations
from typing import Optional, List, Tuple, Literal
from pathlib import Path

import numpy as np
import matplotlib.pyplot as plt
from Bio.Seq import Seq
import pandas as pd

from silvio import (
    DATADIR, # config
    Experiment, ExperimentException, Host, HostException, # base classes
    alldef, coalesce, first, Generator, # utilities
    DataOutcome, DataWithPlotOutcome, combine_data, # outcome
    InsertGeneEvent, # events
    GenomeList, GenomeExpression, GrowthBehaviour, # modules
    Gene, # records
)
# additional non-renamed utilities
from silvio.extensions.utils.misc import check_primer_integrity_and_recombination
from silvio.extensions.utils.visual import Help_Progressbar
from silvio.extensions.utils.laboratory import ErrorRate



class RecExperiment (Experiment) :
    """
    Recombinant Expression Experiment with culture growth and genetic recombination.
    Has a budget system to limit bruteforcing.
    """

    budget: int

    suc_rate: float

    # Keep track of how many hosts were created
    host_counter: int



    def __init__ ( self, seed:Optional[int] = None, equipment_investment:int = 0, max_budget:int = 10000 ) :

        if equipment_investment > max_budget :
            raise ExperimentException("Investment cost is higher than maximal budget.")

        super().__init__(seed=seed)
        self.budget = max_budget - equipment_investment
        self.suc_rate = ErrorRate(equipment_investment, max_budget)
        self.host_counter = 0



    def create_host ( self, species:Literal['ecol','pput'], name:Optional[str] ) -> RecHost:
        self.host_counter += 1
        seed = self.rnd_gen.pick_seed() # The experiment provides stable seed generation for hosts.
        new_host = None

        chosen_name = coalesce( name, species + str(self.host_counter) )
        if species == 'ecol' :
            new_host = self.build_ecol_host( name=chosen_name, seed=seed )
        elif species == 'pput' :
            new_host = self.build_ecol_host( name=chosen_name, seed=seed )

        if species is None :
            raise ExperimentException("Invalid species provided. The possible species are: 'ecol', 'pput'")

        self.bind_host(new_host)
        return new_host



    def build_ecol_host ( self, name:str, seed:int ) -> RecHost :
        gen = Generator( seed )
        host = RecHost( name=name, seed=seed )
        host.make(
            opt_growth_temp= gen.pick_integer(25,40), # unit: degree celsius, source: https://application.wiley-vch.de/books/sample/3527335153_c01.pdf
            max_biomass= gen.pick_integer(30,100), # unit: in gDCW/l, source (german): https://www.google.com/url?sa=t&rct=j&q=&esrc=s&source=web&cd=2&cad=rja&uact=8&ved=2ahUKEwjzt_aJ9pzpAhWGiqQKHb1jC6MQFjABegQIAhAB&url=https%3A%2F%2Fwww.repo.uni-hannover.de%2Fbitstream%2Fhandle%2F123456789%2F3512%2FDissertation.pdf%3Fsequence%3D1&usg=AOvVaw2XfGH11P9gK2F2B63mY4IM
            umax = 1, # standard growth rate for easier result check
            infl_prom_str= gen.pick_integer(30,50), # explanation see Plot_ExpressionRate
            species_prom_str= 0.057,
            opt_primer_len= gen.pick_integer(16,28), # unit: nt, source: https://link.springer.com/article/10.1007/s10529-013-1249-8
            regressor_file= DATADIR / 'expression_predictor' / 'Ecol-Promoter-predictor.pkl',
            addparams_file= DATADIR / 'expression_predictor' / 'Ecol-Promoter-AddParams.pkl',
        )
        host.sync()
        return host



    def build_pput_host ( self, name:str, seed:int ) -> RecHost :
        gen = Generator( seed )
        host = RecHost( name=name, seed=seed )
        host.make(
            opt_growth_temp= gen.pick_integer(25,40), # unit: degree celsius, source: https://application.wiley-vch.de/books/sample/3527335153_c01.pdf
            max_biomass= gen.pick_integer(45,145), # unit: in gDCW/l, source 1: https://onlinelibrary.wiley.com/doi/pdf/10.1002/bit.25474, source 2: https://link.springer.com/article/10.1385/ABAB:119:1:51
            umax = 1, # standard growth rate for easier result check
            infl_prom_str= gen.pick_integer(30,50), # explanation see Plot_ExpressionRate
            species_prom_str= 0.04,
            opt_primer_len= gen.pick_integer(16,28), # unit: nt, source: https://link.springer.com/article/10.1007/s10529-013-1249-8
            regressor_file= DATADIR / 'expression_predictor' / 'Ptai-Promoter-predictor.pkl',
            addparams_file= DATADIR / 'expression_predictor' / 'Ptai-Promoter-AddParams.pkl',
        )
        host.sync()
        return host



    def find_host_or_abort ( self, host_name:str ) -> RecHost :
        """
        Find a host by a given name or abort with an exception.
        Throws: ExperimentException
        """
        host:Optional[RecHost] = first( self.hosts, lambda h: h.name == host_name )
            # Find the first host matching the name.
        if host is None :
            raise ExperimentException("Experiment has not found the host '{}'".format(host_name))
        return host



    def spend_budget_or_abort ( self, amount:int ) -> None :
        """
        Each operation may use up a certain amount of the budget.
        When the budget is not sufficient those operations will all fail.
        By using error handling via raised exceptions we don't need to check each method for a
        success flag on their return value.
        Throws: ExperimentException
        """
        if amount > self.budget :
            raise ExperimentException("Experiment has surpassed its budget. No more operations are allowed.")
        self.budget -= amount



    def simulate_growth ( self, host_name:str, temps:List[int], wait:float = 0.001 ) -> GrowthOutcome :
        """ Simulate a growth under multiple temperatures and return expected biomasses over time. """
        self.spend_budget_or_abort(100)
        host = self.find_host_or_abort(host_name)
        ( df, pauses ) = host.growth.Make_TempGrowthExp(CultTemps=temps, exp_suc_rate=self.suc_rate)

        # wait = 0.001 # has to be adjusted, waiting time for loading bar
        for pause in pauses :
            loading_time = wait * pause.loading_len
            Help_Progressbar(45, loading_time, pause.exp)

        return GrowthOutcome(value=df)



    def clone_with_recombination ( self, host_name:str, primer:Seq, gene:Gene, tm:int ) -> Tuple[RecHost,str] :
        """
        Clone the host while trying to insert a gene. The Clone might or might not contain
        the added Gene.
        Returns: ( new_host, outcome_message )
        """
        self.spend_budget_or_abort(200)
        host = self.find_host_or_abort(host_name)
        rnd = host.make_generator()

        # A clone is always made.
        # TODO: Maybe there is a failure to clone and the return should be `Optional[RecHost]`
        new_host = RecHost( ref=host ) # Clone the host but with different seed.
        new_host.copy( ref=host )
        self.bind_host(new_host)
        if rnd.pick_uniform(0,1) < self.suc_rate: # experiment failure depending on investment to equipment
            return( new_host, 'Cloning failed: Bad Equipment' )

        primer_integrity = check_primer_integrity_and_recombination(
            gene.prom, primer, tm, RecHost.ref_prom, host.genexpr.opt_primer_len
        )
        if not primer_integrity.value :
            return ( new_host, "Primer Failed: " + primer_integrity.error )

        # Both primer integrity and recombination succeeded. Insert the new gene into the clone.
        new_host.insert_gene(gene)
        return ( new_host, "Cloning with recombination succeeded." )



    def measure_promoter_strength ( self, host_name:str, gene_name:str ) -> DataOutcome :
        """
        Measure the promoter strength of gene `gene_name` inside host `host_name`.
        """
        self.spend_budget_or_abort(100)
        host = self.find_host_or_abort(host_name)

        # Calculate the cumulative promoter strength of all genes with the same name.
        matching_genes = [g for g in host.genome.genes if g.name == gene_name]
        prom_str = 0
        for mg in matching_genes :
            prom_str += host.genexpr.calc_prom_str( gene=mg, ref_prom=RecHost.ref_prom )

        return DataOutcome( value=pd.Series({
            'Host': host.name,
            'GeneName': gene_name,
            'GenePromoter': str(matching_genes[0].prom if len(matching_genes) > 0 else None),
                # TODO: Only the promoter from the first matching gene is output. This might be
                #   slightly incorrect if multiple genes with same name have different promoter seqs.
            'PromoterStrength': prom_str
        }) )



    def simulate_vaccine_production (
        self, host_names:List[str], gene_name:Gene, cult_temp:int, growth_rate:float, biomass:int
    ) -> DataOutcome :
        """
        Simulate the protein production of multiple host cultures.
        """
        self.spend_budget_or_abort(500)

        # Equipment failure can prematurely end the simulation.
        if self.rnd_gen.pick_uniform(0,1) < self.suc_rate :
            return DataOutcome( None, 'Experiment failed, bad equipment.' )

        all_outcomes:List[DataOutcome] = []
        for host_name in host_names :

            # Skip this host if it does not exist.
            host = first( self.hosts, lambda h: h.name == host_name )
            if host is None :
                outcome = DataOutcome( pd.Series({ 'Host': host_name }), 'Host does not exist.' )
                all_outcomes.append(outcome)
                continue

            # Skip this host if no matching gene is found.
            gene = first( host.genome.genes, lambda g: g.name == gene_name )
            if gene is None :
                outcome = DataOutcome( pd.Series({ 'Host': host.name, 'Gene_Name': gene_name }), 'Gene does not exist.' )
                all_outcomes.append(outcome)
                continue

            # Simulate the production for this host.
            production = host.growth.Make_ProductionExperiment(
                gene=gene,
                CultTemp=cult_temp,
                GrowthRate=growth_rate,
                Biomass=biomass,
                ref_prom=RecHost.ref_prom,
                accuracy_Test=.9
            )
            outcome = DataOutcome(
                value=pd.Series({
                    'Host': host.name,
                    'Gene_Name': gene.name,
                    'Promoter_Sequence': str(gene.prom),
                    'Promoter_GC-content': (gene.prom.count('C') + gene.prom.count('G')) / len(gene.prom),
                    'Expression_Temperature': cult_temp,
                    'Expression_Biomass': biomass,
                    'Expression_Rate': production.value,
                }),
                error=production.error
            )
            all_outcomes.append(outcome)

        return combine_data(all_outcomes)



    def print_status ( self ) -> None :
        print("Experiment:")
        print("  budget = {}".format( self.budget ))
        print("  failure rate = {}".format( round(self.suc_rate, 2) ))
        print("  hosts = [ {} ]".format( " , ".join([h.name for h in self.hosts]) ))
        # Could display the status of each host if wanted.



class RecHost (Host) :

    growth: GrowthBehaviour

    genome: GenomeList

    genexpr: GenomeExpression

    ref_prom:str = 'GCCCATTGACAAGGCTCTCGCGGCCAGGTATAATTGCACG'



    def make ( self,
        opt_growth_temp:int, max_biomass:int, umax:float,
        infl_prom_str:int, species_prom_str:int, opt_primer_len:int,
        regressor_file:Path, addparams_file:Path
    ) -> None :

        if not alldef( opt_growth_temp, max_biomass, infl_prom_str, species_prom_str, opt_primer_len, regressor_file, addparams_file ) :
            raise HostException("Host not initialized. Reason: incomplete arguments.")

        # Setup GenomeList module
        self.genome = GenomeList()
        self.genome.make()
        self.genome.bind( host=self )

        # Setup GenomeExpression module
        self.genexpr = GenomeExpression()
        self.genexpr.make(
            opt_primer_len=opt_primer_len,
            infl_prom_str=infl_prom_str,
            species_prom_str=species_prom_str,
            regressor_file=regressor_file,
            addparams_file=addparams_file
        )
        self.genexpr.bind( host=self, genome=self.genome )

        # Setup GrowthBehaviour module
        self.growth = GrowthBehaviour()
        self.growth.make(
            opt_growth_temp=opt_growth_temp,
            max_biomass=max_biomass,
            umax = umax
        )
        self.growth.bind( host=self, genexpr=self.genexpr )



    def copy ( self, ref:RecHost ) -> None :

        # Setup GenomeList module using the ref
        self.genome = GenomeList()
        self.genome.copy( ref=ref.genome )
        self.genome.bind( host=self )

        # Setup GenomeExpression module using the ref
        self.genexpr = GenomeExpression()
        self.genexpr.copy( ref=ref.genexpr )
        self.genexpr.bind( host=self, genome=self.genome )

        # Setup GrowthBehaviour module using the ref
        self.growth = GrowthBehaviour()
        self.growth.copy( ref=ref.growth )
        self.growth.bind( host=self, genexpr=self.genexpr )



    def sync ( self ) -> None :
        self.sync_modules([ self.genome, self.genexpr, self.growth ])



    def find_gene_or_abort ( self, gene_name:str ) -> Gene :
        """
        Find a gene by a given name or abort with an exception.
        Throws: HostException
        """
        gene:Optional[Gene] = first( self.genome.genes, lambda g: g.name == gene_name )
            # Find the first gene matching the name.
        if gene is None :
            raise HostException("Host has not found the gene '{}'".format(gene_name))
        return gene



    def insert_gene ( self, gene:Gene ) -> None :
        self.emit( InsertGeneEvent( gene=gene, locus=0 ) )



    def print_status ( self ) -> None :
        print("Host [{}]:".format( self.name ))
        print("  seed plus counter = {} + {}".format( self.rnd_seed, self.rnd_counter ))
        print("  optimal growth temperature = {}".format( self.growth.opt_growth_temp ))
        print("  max biomass = {}".format( self.growth.max_biomass ))
        print("  optimal primer length = {}".format( self.genexpr.opt_primer_len ))
        print("  Gene List: {} genes".format(len(self.genome.genes)))
        for gene in self.genome.genes :
            print("  - {} = {} * {}".format(gene.name, gene.prom, gene.orf))
        print("  Event History: {} events".format(len(self.event_log)))
        for el in self.event_log :
            print("  - {}".format(el))



class GrowthOutcome ( DataWithPlotOutcome ) :

    def make_plot ( self ) -> plt.Figure :
        """
        Plotting with pyplot is unfortunately unintuitive. You cannot display a single figure by
        using the object-oriented API. When you do `plt.subplots` (or create a plot by any other
        means) it will be stored in the global context. You can only display things from the glbbal
        context, and displaying it will remove it from there.
        """
        Time, Biomass = self.value.iloc[:,0], self.value.iloc[:,1:]
        LnBiomass = np.log(Biomass)

        fig, ax = plt.subplots()
        for Exp,X in LnBiomass.iteritems() :
            ax.scatter(Time, X, label=Exp)
        ax.legend()
        return fig




from abc import ABC, abstractmethod
from typing import List, NamedTuple, Dict, Any, Optional
from math import inf

import numpy as np
from Bio import pairwise2
from Bio.Seq import Seq
from Bio.Blast.Record import Alignment

from ....utils import coalesce
from ....tool import Tool, ToolException
from ....random import Generator
from ...common import Base
from .datatype import (
    Scaffold, LocalizedSequence, EstimatedSequence,
    get_consensus_from_overlap, estimate_from_overlap
)




class ContigAssembler (Tool, ABC) :
    """
    Abstract class that defines the interface of all assemblers.
    """

    def __init__ ( self, name:str, seed:Optional[int] = None ) :
        Tool.__init__( self, name=name, seed=seed )

    @abstractmethod
    def apply_internal ( self, scaffolds:List[Scaffold] ) -> List[LocalizedSequence] :
        pass

    @abstractmethod
    def apply ( self, scaffolds:List[Scaffold] ) -> EstimatedSequence :
        pass



class PairwiseScore (NamedTuple) :
    i : int
    k : int
    score : float



def match ( sa:Seq, sb:Seq ) -> Alignment :
    """ Internal method for pairwise matching. """
    return pairwise2.align.localms( sa, sb, 1, -1, -100, -100, gap_char='-' )



class GreedyContigAssembler (ContigAssembler) :
    """
    This assembler will try to assemble the DNA scaffolds by using greedy pairwise matching.
    It will start with the pair of sequences with highest pairwise matching and from there one
    iteratively add one of the remaining sequences by matching it with the consensus sequence of
    all already clustered sequences.

    Attention: It will extract both sequences of paired-end reads, but is not able to use their
    pairing relationship during assembly.
    """

    def __init__ ( self, name:Optional[str] = None, seed:Optional[int] = None ) :
        super().__init__( name=coalesce(name,"GreedyAssembler"), seed=seed )



    def apply_internal ( self, scaffolds:List[Scaffold] ) -> List[LocalizedSequence] :

        # Extract all sequences from the scaffolds. If its 'paired-end' then we also extract the
        # R2 sequences but discard their paired relationship.
        sequences:List[Seq] = []
        for scaf in scaffolds :
            sequences.append( scaf.r1_seqrecord.seq )
            if scaf.r2_seqrecord is not None :
                sequences.append( scaf.r2_seqrecord.seq.reverse_complement().complement() )
                    # Strange how there is no simple `Seq.reverse()` in biopython.

        # Calculate best pairwise matching score to decide on the first sequences to match.
        fbest = PairwiseScore( 0, 0, -inf )
        for i in range(len(sequences)) :
            for k in range( i+1, len(sequences) ) :
                cur_align = match( sequences[i], sequences[k] )[0]
                if cur_align.score > fbest.score :
                    fbest = PairwiseScore( i, 0, cur_align.score ) # 0 because k is not used.

        # To simplify coding (at the expense of an additional loop cycle) we will select the fbest.i
        # as the starting sequence and then recalculate the best pairwise match. It should lead to
        # fbest.k again, but we avoid having to write an extra case for this first match.

        # Keep track of all sequences that are left to cluster.
        available:List[bool] = [ True for _ in range(len(sequences)) ]
        cluster:List[LocalizedSequence] = []

        # Add the first sequence to start the cluster.
        available[fbest.i] = False
        cluster.append( LocalizedSequence( sequence=sequences[fbest.i], locus=0 ) )
        #print("First: " + "".join(sequences[fbest.i]))

        # While sequences are still left unclustered.
        while sum(available) > 0 :
            min_start = min([ locseq.locus for locseq in cluster ])
            consensus = get_consensus_from_overlap(cluster).sequence
            #print("Consensus: " + "".join(consensus))
            best = PairwiseScore( 0, 0, -inf )
            best_align:Alignment = None

            # Get the best matching for the cluster consensus.
            for k in range(len(sequences)) :
                if available[k] :
                    cur_align = match( consensus, sequences[k] )[0]
                    #print("      Try: {} ({}) [{}]".format( "".join(sequences[k]), cur_align.score, k ))
                    #print("         : {}".format("".join(cur_align.seqA)))
                    #print("         : {}".format("".join(cur_align.seqB)))

                    if cur_align.score > best.score :
                        best = PairwiseScore( 0, k, cur_align.score ) # 0 because i is not used.
                        best_align = cur_align

            # Somehow break the entire method if this is None, should never happen.
            if best_align is None :
                raise ToolException("GreedyContigAssembler is not working as intended. Check the code.")

            # From the alignment it is not known if the sequence A starts before or after the
            # sequence B. Unfortunately the `best_align.start` is always positive. Therefore peek
            # which of the sequences starts with the gap character. The one that does will start
            # after the other sequence. All locus is fixed to the locus of the cluster.
            best_locus = min_start
            if best_align.seqB[0] == '-' : # new sequence starts after cluster.
                best_locus = min_start + best_align.start
            elif best_align.seqA[0] == '-' : # new sequence starts before the cluster
                best_locus = min_start - best_align.start
            else : # new sequence start at the same position as the cluster
                best_locus = min_start

            cluster.append( LocalizedSequence( sequence=sequences[best.k], locus=best_locus ))
            available[best.k] = False
            #print("    Added: {} [{}] <{},{}>".format(
            #    "".join(sequences[best.k]), best.k, min_start, best_locus
            #))

        return cluster



    def apply ( self, scaffolds:List[Scaffold] ) -> EstimatedSequence :
        return estimate_from_overlap( self.apply_internal(scaffolds) )



class RandomContigAssembler (ContigAssembler) :
    """
    The random assembler will place all contigs in the scaffolds in random locations.
    Good assemblers should aspire to at least be better than this random assembler.
    """

    expected_genome_size : int



    def __init__ ( self, expected_genome_size:int, name:Optional[str] = None, seed:Optional[int] = None ) :
        self.expected_genome_size = expected_genome_size
        super().__init__( name=coalesce(name,"RandomAssembler"), seed=seed )



    def apply_internal ( self, scaffolds:List[Scaffold] ) -> List[LocalizedSequence] :
        rnd = self.make_generator()

        # Extract all sequences from the scaffolds. If its 'paired-end' then we also extract the
        # R2 sequences but discard their paired relationship.
        sequences:List[Seq] = []
        for scaf in scaffolds :
            sequences.append( scaf.r1_seqrecord.seq )
            if scaf.r2_seqrecord is not None :
                sequences.append( scaf.r2_seqrecord.seq.reverse_complement().complement() )
                    # Strange how there is no simple `Seq.reverse()` in biopython.

        # Randomly decide the start position of each sequence.
        loc_sequences:List[LocalizedSequence] = []
        for seq in sequences :
            start_pos = rnd.pick_integer( 0, self.expected_genome_size - len(seq) )
            loc_sequences.append( LocalizedSequence( sequence=seq, locus=start_pos ) )

        return loc_sequences



    def apply ( self, scaffolds:List[Scaffold] ) -> EstimatedSequence :
        """
        Attributes
        ----------
        scaffolds
            Scaffolds that were returned from fragmentation.

        Returns
        -------
        An estimation of bases for each position, based on random placement.
        """
        return estimate_from_overlap( self.apply_internal(scaffolds) )



"""
Experiment is the top-most scope. It includes all hosts and registries and acts as a global
namespace.
"""

import numpy as np
from typing import Set, Optional, Union
from abc import ABC
import pandas as pd

from numpy.random import SeedSequence

from .host import Host
from .random import Generator



class ExperimentException (Exception) :
    """ Exception that is triggered by an experiment. """
    pass



class Experiment (ABC) :
    """ The basic experiment will hold the hosts and a random-generator. """

    hosts: Set[Host]

    # The experiment will hold an internal random generator to allow repeatability.
    rnd_gen: Generator



    def __init__ ( self, seed:Optional[Union[int,SeedSequence]] = None ) :
        self.hosts = set()
        self.rnd_gen = Generator(seed)



    def bind_host ( self, host:Host ) -> None :
        """Keep the host accessible via the experiment."""
        self.hosts.add(host)



    # def create_host ( self, host_class:Type[Host], **kwargs ) -> None :
    #     host = host_class( exp=self, **kwargs )
    #     self.hosts.append( host )

class ExperimentSettings (ABC):
    """Class to hold experiment settings and parameters."""

    # Dictionary to store chromatograms for time point
    Chromatograms: pd.DataFrame


    # definition of Experiment class with experiment settings
    Test: bool = False # flag to indicate if the experiment is a test
    ExperimentID: str # unique identifier for the experiment
    ExperimentType: str # e.g., "Batch", "Fed-Batch", "Continuous"
    Analytics: list # list of analytics to be performed, e.g., ['OD600', 'HPLC']
    HostName: str # name of the host organism
    Temperature: list # values in Celsius
    CultivationTime: int # default value in hours
    SamplingInterval: float # default sampling interval in hours
    AnalyticSampling: list # specific time points for analytics
    InitBiomass: float # default initial biomass concentration in g/L
    MediumVolume: int # default value in mL
    MediumComposition: dict # key: name of the component, value: concentration in mM
    # CarbonID: str # identifier for the main carbon substrate, following BiGG nomenclature in the GSM model
    CarbonName: list # names of the main carbon substrate
    CarbonSubConc: dict # key: name, as exchange reaction id in model, value: concentration of the carbon substrate in mM
    CarbonUptakeRate: dict # key: name, as exchange reaction id in model, value: uptake rate of the carbon substrate in mmol/gCDW/h
    CarbonSolubleExchange: dict # key: name, carbon exchange reaction id in model, value: exchange rate of carbon metabolite in mmol/gCDW/h (-: uptake, +: secretion)
    CarbonConcDynamic: list # dynamic concentration profile of the carbon substrate over time
    GrowthRate: float # specific growth rate in 1/h
    Yield: dict # key: exchange reactionid, value: biomass yields for soluble metabolites in gCDW/mmol
    Capacity: int # maximum biomass concentration in g/L
    ExchangeRates: dict # key: name, as exchange reaction id in model, value: exchange rate of the metabolite in mmol/gCDW/h
    Results: str # file path to save experiment results

    def __init__(self, Test: bool=False) -> None :
        # initializing the experiment settings with empty values
        self.Test = Test
        self.ExperimentID = ""
        self.ExperimentType = ""
        self.Analytics = []
        self.HostName = ""
        self.Temperature = []
        self.CultivationTime = 0
        self.SamplingInterval = 1.0
        self.AnalyticSampling = []
        self.InitBiomass = 0.0
        self.MediumVolume = 0
        self.MediumComposition = {}
        # self.CarbonID = ""
        self.CarbonName = []  # names of the main carbon substrate
        self.CarbonSubConc = {}
        self.CarbonUptakeRate = {}
        self.CarbonSolubleExchange = {}
        self.CarbonConcDynamic = []
        self.GrowthRate = 0.0
        self.Yield = {}
        self.Capacity = 0
        self.ExchangeRates = {}
        self.Results = ""

        if self.Test:
            self.HostName = "E.coli-core"
            self.InitBiomass = 0.1  # g/L
            # self.CarbonName = ['EX_glc__D_e']
            self.CarbonSubConc = {'EX_glc__D_e': 10.0}  # mM
            self.SamplingInterval = .5
            self.CultivationTime = 10.0

    def set_SamplingVector(self) -> None :
        """Method to generate the sampling vector based on CultivationTime and SamplingInterval."""
        self.SampleVector = np.arange(0, self.CultivationTime + self.SamplingInterval, self.SamplingInterval)

    # the chromatograms dictionary will hold the chromatograms for each time with analytics
    def add_Chromatograms(self, Chromatograms: pd.DataFrame) -> None :
        self.Chromatograms = Chromatograms


"""
Contains various utility methods.
"""

from typing import Any, TypeVar, List, Callable, Optional

T = TypeVar('T')



def coalesce ( *arg ) -> Any :
    """
    Return the first non-null argument given. Emulates a nullish coalesce operator (PEP 505).
    """
    return next( (a for a in arg if a is not None), None )



def alldef ( *arg ) -> bool :
    """
    Returns true if all arguments are non-None.
    """
    return all([ a is not None for a in arg ])



def first ( array:List[T], criteria:Callable[[T],bool] ) -> Optional[T] :
    """
    Return the first item that validates the criteria.
    """
    return next( (item for item in array if criteria(item)), None )


"""
GenomeList is a module that stores the multiple genes an Host may have.

TODO: Add support for adding the same gene twice (probably multisets)
"""

from __future__ import annotations
from copy import copy

from ...host import Host
from ...module import Module
from ...events import EventEmitter, EventLogger
from ..records.gene.gene import Gene
from ..all_events import InsertGeneEvent, RemoveGeneEvent, AlterGenePromoterEvent



class GenomeList ( Module ) :

    genes: set[Gene]



    def make ( self, genes: set[Gene] = set() ) -> None :
        self.genes = genes



    def copy ( self, ref:'GenomeList' ) -> None :
        self.genes = ref.genes.copy()



    def bind ( self, host:Host ) -> None :
        host.observe( InsertGeneEvent, self.listen_insert_gene )
        host.observe( RemoveGeneEvent, self.listen_remove_gene )
        host.observe( AlterGenePromoterEvent, self.listen_alter_gene_promoter )



    def sync ( self, emit:EventEmitter, log:EventLogger ) -> None :
        pass # Nothing to sync.



    def listen_insert_gene ( self, event:InsertGeneEvent, emit:EventEmitter, log:EventLogger ) -> None :
        self.genes.add( event.gene )
        log( "GenomeLibrary: added gene={}".format(event.gene.name) )



    def listen_remove_gene ( self, event:RemoveGeneEvent, emit:EventEmitter, log:EventLogger ) -> None :
        self.genes.remove( event.gene )
        log( "GenomeLibrary: removed gene={}".format(event.gene.name) )



    def listen_alter_gene_promoter ( self, event:AlterGenePromoterEvent, emit:EventEmitter, log:EventLogger ) -> None :
        """ When a gene promoter is altered, we recalculate the expression and update it. """
        found_gene = first( self.locgenes, lambda lg : lg == event.gene )
        if found_gene is not None :
            self.alter_sequence( found_gene.start_loc, found_gene.prom_len, event.new_promoter )
            log( "GenomeLibrary: Changed promoter of gene={} in sequence.".format(event.gene.name) )



    def get_genes_by_name ( self, name:str ) -> List[Gene] :
        found_genes = [ gene for gene in self.locgenes if name == gene.name ]
        return found_genes



"""
Common datatypes between the different extensions.
"""

from typing import List, Literal, Optional


# One of the simple 4 nucleic bases, plus the unknown base.
Base = Literal['A','C','G','T']


# Well distinguished promoter sites.
PromoterSite = Literal[ '-10', '-35' ]


# Read methods usable by the sequencer.
ReadMethod = Literal[ 'single-read', 'paired-end' ]


# BIGG organism dictionary for model download
BIGG_dict = {'E.coli-core':'e_coli_core', 'E.coli': 'iML1515','B.subtilis': 'iYO844', 'P.putida': 'iJN1463', 'S.cerevisiae': 'iMM904'}

# standard units of ExperimentSettings variables
myUnits = {
    'Temperature': 'Celsius',
    'CultivationTime': 'hours',
    'SamplingInterval': 'hours',
    'InitBiomass': 'OD600',
    'MediumVolume': 'mL',
    'CarbonConc': 'mM',
}

# Carbon containing metabolites that are gases, they are treated differently for analytics
GasCarbMets = ['co2', 'ch4']

# Dictionary with analytics and their prices relative to full budget
AnalyticsCosts = {
    'OD600': 0.1,
    'Carbon-Substrate': 0.2,
    'Metabolites': 0.5,
    'pH': 0.1,
    'Biomass': 0.3,
    'HPLC': 0.7,
}

# Dictionary with parameters for unit testing
TestParams = {
    'Time': 10, # hours
}


"""
Methods used in the Transform step of a ETL pipeline.
"""



def list_integer(SeqList):
    '''define input values'''
    alphabet = 'ACGT'
    char_to_int = dict((c,i) for i,c in enumerate(alphabet))
    IntegerList = list()
    for mySeq in SeqList:
        # integer encode input data
        integer_encoded = [char_to_int[char] for char in mySeq.upper()]
        IntegerList.append(integer_encoded)
    return IntegerList



def list_onehot(IntegerList):
    OneHotList = list()
    for integer_encoded in IntegerList:
        onehot_encoded = list()
        for value in integer_encoded:
            letter = [0 for _ in range(4)]
            letter[value] = 1
            onehot_encoded.append(letter)
        OneHotList.append(onehot_encoded)
    return OneHotList



from typing import Set

import numpy as np

from silvio.extensions.utils.misc import (
    Help_simulateHPLCBackground, Help_simulateHPLCTarget, Help_addHPLCPeaks, Help_mergeHPLCChromatograms
)

class SubstrateSideProducts:
    def __init__(self, side_products_dict):
        # side_products_dict: {side_product_name: fractional_flux, ...}
        self.SideProducts = side_products_dict

    def export_dict(self):
        return self.SideProducts


class SideProducts:
    """Class representing side products of metabolic activity."""

    Substrates: Set[str]

    def __init__(self):
        self.Substrates = set()
        # Initialize an empty set of substrates
        # Each substrate will be added as an attribute dynamically


    def add_substrate(self, SubstrateSpectrumDict: dict):
        # SubstrateSpectrumDict: {substrate: {side_product: fractional_flux, ...}, ...}
        # update the set of substrates
        self.Substrates.update(SubstrateSpectrumDict.keys())
        # Dynamically add an attribute for each substrate
        for substrate, side_products in SubstrateSpectrumDict.items():
            setattr(self, substrate, SubstrateSideProducts(side_products))

# Example usage:
# SubstrateSpectrumDict = {
#     'a': {'sp1': 0.5, 'sp2': 0.5},
#     'b': {'sp3': 1.0}
# }

# sp = SideProducts(SubstrateSpectrumDict)
# print(sp.Substrates)  # {'a', 'b'}
# print(sp.a.SideProducts)  # {'sp1': 0.5, 'sp2': 0.5}
# print(sp.b.SideProducts)  # {'sp3': 1.0}

    def exportAll_dict(self):
        '''Export the side products information as a nested dictionary. First level keys are substrates, second level keys are side products with their fractional fluxes.'''
        export = {}
        for substrate in self.Substrates:
            side_products_obj = getattr(self, substrate)
            export[substrate] = side_products_obj.SideProducts
        return export
    
class HPLCProperties:
    class MetaboliteProperties:
        def __init__(self, RetentionTime: float):
            self.RetentionTime = RetentionTime

    def __init__(self):
        self.Metabolites = set()

    def add_metabolite(self, Metabolite: str, RetentionTime: float):
        self.Metabolites.add(Metabolite)
        setattr(self, Metabolite, self.MetaboliteProperties(RetentionTime))

    def init_metabolites(self, MetaboliteList: list):
        for Metabolite in MetaboliteList:
            self.add_metabolite(Metabolite, RetentionTime=np.random.uniform(1.0, 20.0))

    def export_dict(self):
        return {
            'Metabolites': list(self.Metabolites),
        }
    def run_HPLC(self, MetaboliteConcentrations: dict, Test:bool=False):
        '''Simulate HPLC chromatogram based on metabolite concentrations.'''
        # MetaboliteConcentrations: {metabolite_name: concentration, ...}
        # Generate a chromatogram with peaks at the retention times of the metabolites
        # time = np.arange(0, 30.0, 0.01)  # 0 to 30 minutes with 0.01 min resolution
        # signal = np.zeros_like(time)

        if Test:
            MetaboliteConcentrations = {'Product1': [1.0,5.0], 'Product2': [0.0,3.0]}
            df_hplc = Help_simulateHPLCBackground(runtime=30, n_peaks=5, seed=42)
            # construct a list of DataFrames for the combined products at each sampling time
            # For each sampling point, combine all products' peaks with the background
            combined_chroms = []
            for i in range(len(MetaboliteConcentrations['Product1'])):
                temp = df_hplc.copy()
                for prod in MetaboliteConcentrations.keys():
                    prod_peak = Help_simulateHPLCTarget(
                        Elutiontime=np.random.uniform(1.0, 20.0),
                        Concentration=MetaboliteConcentrations[prod][i]
                    )
                    temp = Help_addHPLCPeaks(temp, prod_peak)
                combined_chroms.append(temp)

            # Merge all combined chromatograms into a single DataFrame with one column per sample
            df_prod = Help_mergeHPLCChromatograms(combined_chroms)
            return df_prod
        # for metabolite, concentration in MetaboliteConcentrations.items():
        #     if metabolite in self.Metabolites:
        #         rt = getattr(self, metabolite).RetentionTime
        #         peak = concentration * np.exp(-0.5 * ((time - rt) / 0.1) ** 2)  # Gaussian peak
        #         signal += peak

        # return {
        #     'time': time,
        #     'signal': signal
        # }


"""
The Growth record stores a possible growth outcome.

TODO: Currently not using as GrowthBehaviour.grow is outputting a DataFrame.
  Think about it, but I believe using Records (and maybe Registries) could be better.
  For example, a GrowthRegistry could be a specialized table that can be queried for
  aggregated records or plots.
"""

from __future__ import annotations
from abc import ABC


from ...record import Record



class GrowthOutcome ( Record, ABC ) :

    def __init__ ( self ) :
        super().__init__()

    def clone ( self ) -> GrowthOutcome :
        return GrowthOutcome()



"""
All operations that generate random numbers may come from this globally defined random generator.

This allows the user to provide a seed and generate the same sequences on every run as long as
requests are made in the same order.

The methods it exposes are the same of those from a numpy Generator object, but with typing and
slightly altered parameters.
https://numpy.org/doc/stable/reference/random/generator.html#numpy.random.Generator
"""

import sys
from typing import Any, List, Optional, TypeVar

from numpy.random import Generator as NumpyGenerator, PCG64



T = TypeVar('T')      # Declare type variable



# Initialize the global generator with a random seed first.
_randgen = NumpyGenerator(PCG64())



def set_seed ( new_seed:int ) -> None :
    """
    Sets the seed of the global random generator so that all future random picks follow the same
    pseudo-random sequence.
    """
    global _randgen
    _randgen = NumpyGenerator(PCG64(new_seed))



def pick_choices ( choices:List[T], amount:int, weights:List[float] = None ) -> List[T] :
    """
    Pick a single value out of a provided list. Can receive a non-normalized weights.
    """
    probs = None
    if weights is not None :
        sum_w = sum(list(weights))
        probs = [ w / sum_w for w in list(weights) ]
    return list(_randgen.choice( choices, size=amount, p=probs ))


def pick_samples ( choices:List[T], amount:int ) -> List[T] :
    """
    Pick a sample (without repetition) from choices.
    """
    return list(_randgen.choice( choices, size=amount, replace=False ))


def pick_integer ( low:int, high:int ) -> int :
    """
    Pick a single integer from a range from low (inclusive) to high (exclusive).
    """
    return int(_randgen.integers( low, high, endpoint=False ))


def pick_uniform ( low:float = 0, high:float = 1 ) -> float :
    """
    Pick a single float from an uniform distribution between low (inclusive) and high (exclusive).
    """
    return float(_randgen.uniform( low, high ))


def pick_normal ( mean:float = 0, sd:float = 0 ) -> float :
    """
    Pick a single number from a normal distribution.
    """
    return float(_randgen.normal( mean, sd ))


def pick_exponential ( beta:float ) -> float :
    """
    Pick a single number from an exponential distribution.
    """
    return float(_randgen.exponential( beta ))


def pick_seed () -> int :
    """
    Pick a new number that is usable as a seed.
    """
    return int(_randgen.integers(sys.maxsize))



class Generator () :
    """ All random methods also exist inside a local generator object. """

    gen: NumpyGenerator

    def __init__ ( self, seed:Optional[int] = None ) :
        self.gen = NumpyGenerator(PCG64(seed))

    def pick_choices ( self, choices:List[T], amount:int, weights:List[float] = None ) -> List[T] :
        probs = None
        if weights is not None :
            sum_w = sum(list(weights))
            probs = [ w / sum_w for w in list(weights) ]
        return list(self.gen.choice( choices, size=amount, p=probs ))

    def pick_samples ( self, choices:List[T], amount:int ) -> List[T] :
        return list(self.gen.choice( choices, size=amount, replace=False ))

    def pick_integer ( self, low:int, high:int ) -> int :
        return int(self.gen.integers( low, high, endpoint=False ))

    def pick_uniform ( self, low:float = 0, high:float = 1 ) -> float :
        return float(self.gen.uniform( low, high ))

    def pick_normal ( self, mean:float = 0, sd:float = 0 ) -> float :
        return float(self.gen.normal( mean, sd ))

    def pick_exponential ( self, beta:float ) -> float :
        return float(self.gen.exponential( beta ))

    def pick_seed ( self ) -> int :
        return int(self.gen.integers(sys.maxsize))



"""
StubGene is a gene with no information about its actual sequence.
It only has properties about its name.
"""

from __future__ import annotations

from Bio.Seq import Seq

from .gene import Gene



class StubGene (Gene) :

    _name: str

    def __init__ ( self, name:str ) :
        super().__init__()
        self._name = name

    @property
    def name ( self ) -> str :
        return self._name

    @property
    def orf ( self ) -> Seq :
        return Seq("Z")

    @property
    def prom ( self ) -> Seq :
        return Seq("Z")



"""
An experiment that uses hosts which implements all currently existing modules.
"""

from __future__ import annotations
from typing import Optional, List, Tuple, Literal
from pathlib import Path

import os
import copy
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import scipy.integrate as spi
from cobra.core.model import Model


from silvio import (
    DATADIR, # config
    Experiment, ExperimentException, Host, HostException, ExperimentSettings, # base classes
    alldef, coalesce, first, Generator, # utilities
    DataOutcome, DataWithPlotOutcome, combine_data, add_dfcolumn, # outcome
    GrowthBehaviour, MetabolicFlux, # modules
)
# additional non-renamed utilities
from silvio.extensions.utils.visual import Help_Progressbar
from silvio.extensions.utils.laboratory import ErrorRate
from silvio.extensions.utils.misc import Help_GrowthConstant, Download_GSMM, add_noise
from silvio.extensions.modules.metabolism import HPLCProperties


class DigLabSim (Experiment) :
    """
    Growth experiments on Monod equation with temperature, substrate and sampling times.
    """

    budget: int

    suc_rate: float

    # Keep track of how many hosts were created
    host_counter: int

    # storing HPLC properties
    HPLC = HPLCProperties()

    # Record of all experiments conducted
    class experiment_history:
        pass

    # # Store experiment settings by ExperimentID for easy access
    # experiment_history: dict[str, ExperimentSettings] = {}


    def __init__ ( self, seed:Optional[int] = None, equipment_investment:int = 0, max_budget:int = 10000  ) :

        if equipment_investment > max_budget :
            raise ExperimentException("Investment cost is higher than maximal budget.")

        super().__init__(seed=seed)
        self.budget = max_budget - equipment_investment
        self.suc_rate = ErrorRate(equipment_investment, max_budget)
        self.host_counter = 0
        self.HPLC = HPLCProperties()
        self.experiment_history = {} # initialize empty history dict


    def create_host ( self, name:Optional[str] ) -> GroHost:
        self.host_counter += 1
        seed = self.rnd_gen.pick_seed() # The experiment provides stable seed generation for hosts.
        new_host = None

        new_host = self.build_host( name=name, seed=seed )

        self.bind_host(new_host)
        return new_host



    def build_host ( self, name:str, seed:int ) -> GroHost :
        gen = Generator( seed )
        host = GroHost( name=name, seed=seed )
        GSMM = Download_GSMM(name)
        Yxs = round(gen.pick_uniform(.15, .55), 2) # unit: gDW/gSubstrate, source: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4681042/
        opt_growth_temp= gen.pick_integer(25, 40) # unit: degree celsius, source: https://application.wiley-vch.de/books/sample/3527335153_c01.pdf
        max_biomass= gen.pick_integer(30, 100) # unit: in gDW/L, source (german): https://www.google.com/url?sa=t&rct=j&q=&esrc=s&source=web&cd=2&cad=rja&uact=8&ved=2ahUKEwjzt_aJ9pzpAhWGiqQKHb1jC6MQFjABegQIAhAB&url=https%3A%2F%2Fwww.repo.uni-hannover.de%2Fbitstream%2Fhandle%2F123456789%2F3512%2FDissertation.pdf%3Fsequence%3D1&usg=AOvVaw2XfGH11P9gK2F2B63mY4IM
        Ks = round(gen.pick_uniform(1, 10), 3) # unit: gSubstrate/L, source: https://bionumbers.hms.harvard.edu/bionumber.aspx?s=n&v=3&id=111049, for K12
        k1 = round(gen.pick_uniform(.05, .2), 3)
        umax = round(gen.pick_uniform(.1, 2*Yxs), 3) # unit: /h, the maximum growth rate determines with the biomass yield the max glucose uptake rate. This should be below 10 mmol/gDW/h
        OD2X = round(gen.pick_uniform(0.3, 0.5), 3)
        host.make(
            opt_growth_temp=opt_growth_temp,
            max_biomass=max_biomass,
            Ks=Ks,
            Yxs=Yxs,
            k1=k1,
            umax=umax,
            OD2X=OD2X,
            GSMM=GSMM
        )
        host.sync()
        
        # adding kinetic information to GSMM
        host.metabolism.set_KinParameters()

        return host


    def find_host_or_abort ( self, host_name:str ) -> GroHost :
        """
        Find a host by a given name or abort with an exception.
        Throws: ExperimentException
        """
        host:Optional[GroHost] = first( self.hosts, lambda h: h.name == host_name )
            # Find the first host matching the name.
        if host is None :
            raise ExperimentException("Experiment has not found the host '{}'".format(host_name))
        return host

    def spend_budget_or_abort ( self, amount:int ) -> None :
        """
        Each operation may use up a certain amount of the budget.
        When the budget is not sufficient those operations will all fail.
        By using error handling via raised exceptions we don't need to check each method for a
        success flag on their return value.
        Throws: ExperimentException
        """
        if amount > self.budget :
            raise ExperimentException("Experiment has surpassed its budget. No more operations are allowed.")
        self.budget -= amount

    def record_experiment(self, settings: ExperimentSettings) -> None:
        """
        Record the details of an experiment along with its settings.
        Args:
            settings (ExperimentSettings): The settings of the experiment to be recorded.
        """
        self.experiment_history[settings.ExperimentID] = settings

    def measure_DryWeight( self, host_name:str, Temperature:float, FinalTime:float, SubstrateConc:float, Replicates: int, FileName: str ) -> DataOutcome :
        '''Experiment to determine the correlation between OD and dry weight. The factor depends on the temperature. Get growth phase when substrate is at least half and not fully consumed.
        Args:
            host_name: name of the host
            Temperature: float of temperature for the experiment
            FinalTime: float of total experiment time in h
            SubstrateConc: float of substrate concentration
            Replicates: int of number of replicates
            FileName: name of the file to which the data is saved

        Returns:
            DataOutcome: dataframe with the expected biomass over time
            CSV file: csv file with the data
        '''

        host = self.find_host_or_abort(host_name)
        # Equipment failure can prematurely end the simulation.
        if self.rnd_gen.pick_uniform(0,1) < self.suc_rate :
            return DataOutcome( None, 'Experiment failed, bad equipment.' )

        # calculating the experiment price based on the number of samples
        BaseCost = 25
        TotalCost = round(BaseCost * Replicates)
        self.spend_budget_or_abort( TotalCost )

        OptTemp = host.growth.opt_growth_temp
        Initials={'X0': 0.01, 'S0': SubstrateConc}
        Params={'mumax': host.growth.umax, 'Ks': host.growth.Ks, 'Yxs': host.growth.Yxs, 'max_biomass': host.growth.max_biomass}
        Params['mumax'] = Params['mumax'] * Help_GrowthConstant(host.growth.opt_growth_temp, Temperature, 4)
        ResAll = pd.DataFrame(columns=['t', 'X', 'S', 'OD'])
        for Rep in range(Replicates):
            # solving the Monod equation and adding noise
            MonodSol = add_noise(solve_MonodEqn(Params, Initials, np.linspace(0,FinalTime,50)), self.suc_rate)
            # Last row of MonodSol combined with FinalTime
            Last = np.hstack((FinalTime, MonodSol[-1,:]))
            Result = round(pd.DataFrame(Last.reshape(1,-1), columns=['t', 'X', 'S']),3)
            # adding OD with deviation from Temperature, the variance is set larger (5) so the effect is not too drastic
            OD2X = host.growth.OD2X * Help_GrowthConstant(host.growth.opt_growth_temp, Temperature, 5)
            Result['OD'] = round(X2OD(Result['X'], OD2X),3)
            # add Result dataframe at the bottom of a new dataframe
            ResAll = pd.concat([ResAll, Result], ignore_index=True)
        DataReturn = DataOutcome(ResAll)
        # save DataReturn to csv
        # relative path to the data folder
        FilePath = os.path.join('..', 'Data', FileName)
        DataReturn.export_data(FilePath)
        return DataReturn

    def measure_TemperatureGrowth( self, ExperimentSettings: ExperimentSettings, Test:str = 'Standard') -> DataOutcome :
        """ 
        Simulate a growth under multiple temperatures and return expected biomasses over time.
        Args:
            ExperimentSettings: settings for the experiment
            Analytics: list of analytics to be performed, e.g., ['OD600', 'HPLC']
            Test: if True, a test experiment will be performed with default settings
        Returns:
            GrowthOutcome: dataframe with the expected biomass over time
        """

        host = self.find_host_or_abort( ExperimentSettings.HostName ) #host_name:str
        if Test == 'Standard':
            # check if money is available
            self.spend_budget_or_abort( len(ExperimentSettings.Temperature)*100 )

            Analytics = ExperimentSettings.Analytics

            # the while loop searches for side products until a stable solution is found
            SideProdsNum = self.rnd_gen.pick_integer(2,5)
            # set the random fraction of side products to the substrate carbon molarity
            SideProdFrac = self.rnd_gen.pick_uniform(0.05, 0.3)
            StableSol = False
            while not StableSol:
            # set uptake rate of selected substrate
                host.metabolism.set_resetCarbonExchanges(ExperimentSettings.CarbonSubConc)
                _, StableSol = host.metabolism.add_ProductSpectrum(CarbSubstrate=ExperimentSettings.CarbonID, CarbSideNumb=SideProdsNum, CarbSideFract=SideProdFrac, Test=Test)

            ExperimentSettings.GrowthRate, ExperimentSettings.ExchangeRates = host.metabolism.optimize_ReportExchanges()
            ExperimentSettings.CarbonSolubleExchange = host.metabolism.get_SolubleProducts(ExperimentSettings.ExchangeRates)
            ExperimentSettings.Yield = {rxn: round(ExperimentSettings.GrowthRate/value,2) for rxn, value in ExperimentSettings.CarbonSolubleExchange.items()} # gCDW/mmol
            # calculate biomass capacity in gCDW/L for yields with negative exchange rates (uptake) in ExperimentSettings.Yield dictionary
            BioWeightConc = sum([ExperimentSettings.Yield[rxn] * ExperimentSettings.CarbonSubConc[rxn] for rxn in ExperimentSettings.Yield.keys() if ExperimentSettings.CarbonSolubleExchange[rxn] < 0])  # gCDW/L
            ExperimentSettings.Capacity = BioWeightConc #* (ExperimentSettings.MediumVolume/1000)  # gCDW in the flask
            # ExperimentSettings.Capacity = CapacityWeight / host.growth.OD2X  # convert gCDW to OD600


            # ExpFail_rand= host.make_generator().pick_uniform(0,1)

            # loop over all temperatures
            for TestTemp in ExperimentSettings.Temperature:
                # copy the experiment settings to avoid overwriting and set the temperature
                ExperimentSettings_Temp = copy.deepcopy(ExperimentSettings)
                ExperimentSettings_Temp.Temperature = TestTemp
                # Equipment failure can prematurely end the simulation.
                # if ExpFail_rand > self.suc_rate :
                    # calculating the biomass in gCDW/L over time
                Biomass = host.growth.Stream_TempGrowthExp( ExperimentSettings=ExperimentSettings_Temp, Success_rate=self.suc_rate )
                # else:
                #     # no growth, values around the initial biomass are selected from a normal distribution
                #     Biomass = np.random.normal( ExperimentSettings.InitBiomass, .01*ExperimentSettings.InitBiomass, len(ExperimentSettings_Temp.SampleVector) )

                # create dataframe with time and biomass concentration in OD
                df = pd.DataFrame({'time (h)': ExperimentSettings.SampleVector, 
                                    'biomass (OD600) at {} °C'.format(TestTemp): Biomass/host.growth.OD2X}) # convert gCDW to OD600

                if 'Biomass' in Analytics:
                    # add dry weight concentration over time if selected in analytics to the existing dataframe
                    df['biomass (gCDW/L) at {} °C'.format(TestTemp)] = Biomass

                if 'Carbon-Substrate' in Analytics:
                    Carbon = host.growth.Stream_CalcSubConc( Biomass=Biomass, ExperimentSettings=ExperimentSettings, Success_rate=self.suc_rate )
                    # add substrate concentration over time if selected in analytics to the existing dataframe
                    df['{} conc. (mM) at {} °C'.format(ExperimentSettings.CarbonName, TestTemp)] = Carbon
                    ExperimentSettings.CarbonConcDynamic = Carbon

                if 'Metabolites' in Analytics:
                    # calculate product spectrum based on carbon uptake rate and biomass concentration
                    myProd = host.metabolism.get_SolubleProducts(ExchRct=ExperimentSettings.ExchangeRates)
                    Spectrum = host.growth.Stream_CalcProdConc(Biomass,myProd,ExperimentSettings.SampleVector,self.suc_rate, Test=Test)
                    # converting Spectrum dict to dataframe to merge with df
                    df_Spectrum = pd.DataFrame(Spectrum)
                    # Set entries to zero that are not sampled in ExperimentSettings.AnalyticSampling
                    df_Spectrum = df_Spectrum.reindex(ExperimentSettings.AnalyticSampling, fill_value=0)
                    # add temperature to the column names
                    df_Spectrum.columns = [col + ' conc. (mM) at {} °C'.format(TestTemp) for col in df_Spectrum.columns]
                    # merge df with df_Spectrum based on time
                    df = pd.merge(df, df_Spectrum, left_on='time (h)', right_index=True, how='left')

                if 'HPLC' in Analytics:
                    self.HPLC.run_HPLC('default',Test=True) 
                    # # calculate chromatogram based on metabolite concentrations
                    # Chromatogram = self.HPLC.calc_Chromatogram(df, ExperimentSettings.AnalyticSampling)
                    # # add temperature to the column names
                    # Chromatogram.columns = [col + ' at {} °C'.format(TestTemp) for col in Chromatogram.columns]
                    # # add chromatogram to the experiment settings
                    # if self.Chromatograms is not None:
                    #     self.Chromatograms = pd.merge(self.Chromatograms, Chromatogram, left_on='time (h)', right_index=True, how='outer')
                    # else:
                    #     self.Chromatograms = Chromatogram

                # add biomass to existing dataframe with previous results
                if 'Result' in locals():
                    Result = pd.merge(Result, df, on='time (h)', how='outer') 
                else:
                    Result = df


        elif Test == 'Example': # general test with example parameters
            StableSol = False
            count = 1
            while not StableSol and count < 15:
                host.metabolism.set_resetCarbonExchanges(ExperimentSettings.CarbonSubConc)
                _, StableSol = host.metabolism.add_ProductSpectrum(CarbSubstrates=['EX_glc__D_e'],CarbSideNumb=3, CarbSideFract=0.15, Test=False)
                count += 1
                if count == 15:
                    print('Max iterations reached without convergence')
                    break

            ExperimentSettings.Temperature = host.opt_growth_temp
            ExperimentSettings.GrowthRate, ExperimentSettings.ExchangeRates = host.metabolism.optimize_ReportExchanges()
            ExperimentSettings.CarbonSolubleExchange = host.metabolism.get_SolubleProducts(ExchRct=ExperimentSettings.ExchangeRates)
            ExperimentSettings.Yield = {rxn: round(ExperimentSettings.GrowthRate/value,2) for rxn, value in ExperimentSettings.CarbonSolubleExchange.items()} # gCDW/mmol
            ExperimentSettings.Capacity = sum([abs(ExperimentSettings.Yield[rxn]) * ExperimentSettings.CarbonSubConc[rxn] for rxn in ExperimentSettings.Yield.keys() if ExperimentSettings.CarbonSolubleExchange[rxn] < 0])  # gCDW/L
            ExperimentSettings.Temperature = host.opt_growth_temp
            Biomass = host.growth.Stream_TempGrowthExp( ExperimentSettings=ExperimentSettings, Success_rate=0 )

            # create dataframe with time and biomass concentration in OD
            df = pd.DataFrame({'time (h)': ExperimentSettings.SampleVector, 
                                'biomass (OD600)': Biomass/host.growth.OD2X}) # convert gCDW to OD600
            df.reset_index(drop=True, inplace=True)

            Spectrum = host.growth.calc_CarbonSolubleConcentrationDynamics(Biomass, ExperimentSettings)
            # converting Spectrum dict to dataframe to merge with df
            df_Spectrum = pd.DataFrame(Spectrum)
            df_Spectrum.reset_index(drop=True, inplace=True)


            # # Set entries to zero that are not sampled in ExperimentSettings.AnalyticSampling
            # df_Spectrum = df_Spectrum.reindex(ExperimentSettings.AnalyticSampling, fill_value=0)
            # merge df with df_Spectrum based on time
            Result = pd.concat([df, df_Spectrum], axis=1)

            # generating HPLC sample data
            # first setting random retention times for all metabolites with init_metabolites method in HPLCProperties class
            hplc_df = self.HPLC.run_HPLC(MetaboliteConcentrations=[], Test=True)
            rt = hplc_df['time']
            signals = hplc_df.drop(columns=['time']).T.values
            Result['HPLC'] = add_dfcolumn(
                Vector1=rt,
                Array=signals,
                RowIndex=[0,20], # just as example, adding two chromatograms at time 0 and 20
                n_rows=len(Result))



        elif Test == 'Unit': # unit test with fixed parameters
            emptyExp = ExperimentSettings
            ExperimentSettings.ExperimentID = "Test"
            host = self.find_host_or_abort( [h.name for h in self.hosts][0] ) # take the first host
            Biomass = host.growth.Stream_TempGrowthExp( ExperimentSettings=emptyExp, Test=True )
            # create dataframe with time and biomass concentration in OD
            df = pd.DataFrame({'time (h)': range(10), # in Test the time is just 0-10
                                'biomass (OD600) at Test': Biomass})

            # add biomass to existing dataframe with previous results
            if 'Result' in locals():
                Result = pd.merge(Result, df, on='time (h)', how='outer') 
            else:
                Result = df
        else:
            raise ExperimentException("Test parameter not recognized.")
            
        # rounding the result to 2 decimals
        Result = round(Result, 2)
        DataOutcome = GrowthOutcome(value=Result)

        # adding experiment settings to the history
        self.record_experiment(ExperimentSettings)

        return DataOutcome


    def measure_BiomassSubstrateExp( self, host_name:str, Temperature:float, Sampling:List[float], SubstrateConc:List[float], NightShift:float, FileName:str, wait:float = 0.01, Function='Monod' ) -> DataOutcome :
        """ Simulate a growth under multiple temperatures and return expected biomasses over time. 
        Args:
            host_name: name of the host
            Temperature: float of temperature for the experiment
            Sampling: array of (total experiment time in h, sampling interval in h) with shape (len(temps),2)
            SubstrateConc: float of substrate concentration
            NightShift: time of the night shift in h after which there are no measurements for 6h
            FileName: name of the file to which the data is saved
            Function: optional, function to be used for the simulation, default is Monod
        
        Returns:
            GrowthOutcome: dataframe with the expected biomass over time
        """
        host = self.find_host_or_abort(host_name)
        # Equipment failure can prematurely end the simulation.
        if self.rnd_gen.pick_uniform(0,1) < self.suc_rate :
            print('Experiment failed, bad equipment.')
            return DataOutcome( None, 'Experiment failed, bad equipment.' )

        # The nightshift has to start within a day of the experiment and must be shorter than 15h
        if NightShift > 15:
            raise ExperimentException("Night shift is too late. It has to be within 15h of the experiment.")
        if Function == 'Test':
            import random
            samples = random.randint(8, 12)
            time = np.linspace(0, 10, samples)
            mu = [.5,.2]
            biomass = [.01*np.exp(mu*time) for mu in mu]
            Result = pd.DataFrame(np.vstack((time,biomass)).T)
            DataReturn = DataOutcome(Result)
            DataPlot = GrowthOutcome(Result)
            DataPlot.make_plot()

        elif Function == 'Monod':
            Params={'mumax': host.growth.umax, 'Ks': host.growth.Ks, 'Yxs': host.growth.Yxs, 'max_biomass': host.growth.max_biomass}
            Initials={'X0': 0.01, 'S0': SubstrateConc}
            time=np.linspace(0, Sampling[0], int(Sampling[0]/Sampling[1]))
            # adding night shift
            NightDuration = 6 # This long is a night without measurements
            time = add_NightShift(time, NightShift, NightDuration)
            # calculating the experiment price based on the number of samples
            BaseCost = 100
            TotalCost = round(BaseCost * calculate_ExpPriceFactor(len(time)))
            self.spend_budget_or_abort( TotalCost )
            # adjusting the real growth rate based on the temperature in the experiment, using variance = 4, which leads to half maximum growth rate for 5C difference
            Params['mumax'] = Params['mumax'] * Help_GrowthConstant(host.growth.opt_growth_temp, Temperature, 4)
            # solving the Monod equation and adding noise
            MonodSol = add_noise(solve_MonodEqn(Params, Initials, time), self.suc_rate)
            Result = round(pd.DataFrame(np.vstack((time, MonodSol.T)).T, columns=['t', 'X', 'S']),3)
            # Converting the dry weight to OD
            Result['OD'] = Result['X'].apply(lambda x: round(X2OD(x, host.growth.OD2X),3))
            # Deleting the column X
            Result = Result.drop(columns=['X'])

            # # Experiments take time...
            # pause = len(time)
            # loading_time = wait * pause
            # Help_Progressbar(45, loading_time, ' experiment')

            DataReturn = DataOutcome(Result)
            # DataPlot = GrowthOutcome(Result)
            # DataPlot.make_plot()

        FilePath = os.path.join('..', 'Data', 'GrowthExperiment_StandardFormat.xlsx')
        sheet = FileName
        DataReturn.append_data2xlsx(FilePath, sheet)
        return DataReturn



    def print_status ( self ) -> None :
        print("Experiment:")
        print("  budget = {}".format( self.budget ))
        print("  failure rate = {}".format( round(self.suc_rate, 2) ))
        print("  hosts = [ {} ]".format( " , ".join([h.name for h in self.hosts]) ))
        # Could display the status of each host if wanted.

    def check_Results( self, host_name:str, Results:dict ) :
        '''Check if the results are correct. Some parameters are individually checked, so the code is a little more complex...'''
        host = self.find_host_or_abort(host_name)
        # Reference dictionary to find the right parameter from Results in host growth
        ParID_dict = {'Temperature':'opt_growth_temp', 
                      'MaxBiomass':'max_biomass',
                      'OD2X':'OD2X',
                      'GrowthRate_Avg':'umax', 
                      'GrowthRate':'umax', 
                      'GrowthYield_Avg':'Yxs',
                      'Ks_Avg':'Ks', 
                      'GlcRateMax':'GlcRateMax'}
        # Setting correct units for each parameter
        Units_dict = {'Temperature':u'\u00b0C', 
                'MaxBiomass':'gDW/L',
                'OD2X':'a.u.',
                'GrowthRate':'/h', 
                'GrowthYield':'g/g',
                'Ks':'g/L', 
                'GlcRateMax':'mmol/gDW/h'}

        # Delete all Results which are set to None
        Results = {k: v for k, v in Results.items() if v is not None}
        # Result comparison when standard deviations are available
        for Parameter in np.unique([Parameter.split('_')[0] for Parameter in Results.keys()]):
            if ''.join([Parameter,'_Std']) in Results.keys() and Parameter != 'GlcRateMax':
                refval = getattr(host.growth, ParID_dict[''.join([Parameter,'_Avg'])]) #vars(host.growth)
                value = Results[''.join([Parameter,'_Avg'])]
                stdev = Results[''.join([Parameter,'_Std'])]
                upper = value + stdev
                lower = value - stdev
            # only one value for the parameter exists, no standard deviation, e.g. Temperature
                if  refval < upper and refval > lower:
                    print(f'{Parameter}: {value}±{stdev} {Units_dict[Parameter]}',u'\u2705')
                else:
                    # calculating the fold of standard deviation between reference value and test value
                    Fold = np.abs(value - refval)/stdev
                    print(f'{Parameter}: {value}±{stdev} {Units_dict[Parameter]}', u'\u274C', f'Value is {Fold:.1f}x standard deviations from the reference value')
            # calculating the solution for the maximum glucose uptake rate which is the quotient of maximum growth rate and yield coefficient
            elif Parameter == 'GlcRateMax':
                refval = getattr(host.growth,'umax')/getattr(host.growth,'Yxs')/.18 # Yxs is given as g/g, but we need mmol/g
                value = Results[''.join([Parameter,'_Avg'])]
                stdev = Results[''.join([Parameter,'_Std'])]
                upper = value + stdev
                lower = value - stdev
            # only one value for the parameter exists, no standard deviation, e.g. Temperature
                if  refval < upper and refval > lower:
                    print(f'{Parameter}: {value}±{stdev} {Units_dict[Parameter]}',u'\u2705')
                else:
                    # calculating the fold of standard deviation between reference value and test value
                    Fold = np.abs(value - refval)/stdev
                    print(f'{Parameter}: {value}±{stdev} {Units_dict[Parameter]}', u'\u274C', f'Value is {Fold:.1f}x standard deviations from the reference value')
            else:
                # calculating ratio of experimental and theoretical value
                Ratio = Results[Parameter]/getattr(host.growth,ParID_dict[Parameter])
                # if Ratio close around 1 then the measurement is right:
                if Ratio < 1.05 and Ratio > .95:
                    print(f'{Parameter}: {Results[Parameter]} {Units_dict[Parameter]}',u'\u2705')
                else:
                    print(f'{Parameter}: {Results[Parameter]} {Units_dict[Parameter]}',u'\u274C')



class GroHost (Host) :

    growth: GrowthBehaviour
    metabolism: MetabolicFlux

    def make ( self, opt_growth_temp:int, max_biomass:int , GSMM:Model, Ks:float, Yxs:float, k1:float, umax:float, OD2X:float ) -> None :

        if not alldef( opt_growth_temp, GSMM, max_biomass, Ks, Yxs, k1, umax, OD2X ) :
            raise HostException("Host not initialized. Reason: incomplete arguments.")

        # Setup GrowthBehaviour module
        self.growth = GrowthBehaviour()
        self.opt_growth_temp = opt_growth_temp
        self.max_biomass = max_biomass


        self.growth.make2(
            opt_growth_temp=opt_growth_temp, max_biomass=max_biomass, Ks=Ks, Yxs=Yxs, k1=k1, umax=umax, OD2X=OD2X
        )

        # Setup Metabolism module
        self.metabolism = MetabolicFlux()
        self.metabolism.bind( host=self )
        self.metabolism.make( model=GSMM )

        self.growth.bind2( host=self )



    def copy ( self, ref:GroHost ) -> None :

        # Setup GrowthBehaviour module using the ref
        self.growth = GrowthBehaviour()
        self.growth.copy( ref=ref.growth )
        self.growth.bind2( host=self )



    def sync ( self ) -> None :
        self.sync_modules([ self.growth ])




    def print_status ( self ) -> None :
        print("Host [{}]:".format( self.name ))
        print("  seed plus counter = {} + {}".format( self.rnd_seed, self.rnd_counter ))
        print("  optimal growth temperature = {}".format( self.growth.opt_growth_temp ))
        print("  max biomass = {}".format( self.growth.max_biomass ))
        print("  Event History: {} events".format(len(self.event_log)))
        for el in self.event_log :
            print("  - {}".format(el))


class GrowthOutcome ( DataWithPlotOutcome ) :

    def make_plot ( self , XName: str, YNames: list) -> plt.Figure :
        """
        Plotting with pyplot is unfortunately unintuitive. You cannot display a single figure by
        using the object-oriented API. When you do `plt.subplots` (or create a plot by any other
        means) it will be stored in the global context. You can only display things from the glbbal
        context, and displaying it will remove it from there.
        """
        x = self.value.loc[:, XName]
        # LnBiomass = np.log(Biomass)

        fig, ax = plt.subplots()
        for YName in YNames: # LnBiomass.iteritems(), Biomass.iteritems()
            ax.scatter(x, self.value.loc[:, YName], label=YName)
        ax.set_xlabel(XName)
        ax.legend()
        return fig, ax

    def make_3dplot ( self, Column: str, Dictx: dict, Dicty: dict ) -> plt.Figure :
        """
        Make a 3D plot of the data in the DataFrame.
        Column: column name to plot on z-axis
        Dictx: dictionary mapping x-axis values to column names
        Dicty: dictionary mapping y-axis values to column names
        """
        from mpl_toolkits.mplot3d import Axes3D  # noqa: F401 unused import

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        Data = self.value.dropna(subset=[Column])[Column]

        for i, entry in enumerate(Data.values):
            x_data = entry[Dictx]
            y_data = entry[Dicty]

            ax.plot(x_data, y_data, i, zdir='y', label=f'HPLC {i}')

        ax.set_xlabel(Dictx)
        ax.set_ylabel(Dicty)
        ax.set_zlabel(Column)

        return fig, ax


def calculate_ExpPriceFactor(SampleAmount, PlotExample=False):
    '''logit function to generate the price-factor for amount of sampling
        https://nathanbrixius.wordpress.com/2016/06/04/functions-i-have-known-logit-and-sigmoid/
        increasing sampling amount will increase the price until a certain point
        after that the price will remain constant until a certain point
        after that the price will increase again
        the input needs to be between 0 and 1, otherwise the function will return an error
        therefore we need to normalize the input with a saturation function
        the saturation function is a tanh function
        the saturation function will be used to normalize the input

        Args:
            SampleAmount (float): amount of sampling
            PlotExample (bool): if True, the function will plot an example of the function
        Returns:
            PriceFactor (float): price factor for the amount of sampling
        '''
    myTanh = lambda x: np.tanh(x/20) #+ np.tanh(1/20)
    myLogit = lambda x: (np.log(x/(1-x))**3 - np.log(myTanh(1)/(1-myTanh(1)))**3 + 1)/26
    PriceFactor = myLogit(myTanh(SampleAmount))
    if PlotExample:
        x = np.linspace(1,SampleAmount,SampleAmount)
        y = myLogit(myTanh(x))
        plt.plot(x,y, label='logit function', marker='o', linestyle='None')
        plt.xlabel('Sampling amount')
        plt.ylabel('Price factor')
        plt.show()

    return PriceFactor


def MonodEqnODE(MonodVars, t, mumax, Ks, Yxs, max_biomass):
    """
    Monod bioprocess model with ODEs. Based on the model described in:
    Hemmerich et al., doi:10.1002/elsc.202000088, Fig. 6

    Args:
        MonodVars (list): list of variables [X, S]
        t (float): time
        mumax (float): maximum specific growth rate
        Ks (float): half-saturation constant, unit: g/L
        Yxs (float): yield coefficient, unit: gDW/gSubstrate

    Returns:
        list: list of derivatives [dX_dt, dS_dt]
    """
    # Unpack y and z
    X, S = MonodVars

    mu = mumax * S / (Ks + S)
    # Compute biomass growth rate with  a Verhulst growth function
    dX_dt = mu * X * (1 - X / max_biomass)

    # Compute substrate consumption rate
    dS_dt = -mu * X  * (1 - X / max_biomass) / Yxs

    # Return the result as a NumPy array
    return np.array([dX_dt, dS_dt])

def solve_MonodEqn(Params, Initials, t):
    """
    Solve the Monod bioprocess model with ODEs.

    Args:
        Params (dict): dictionary of parameters
        Initials (dict): dictionary of initial conditions
        t (list): list of time points

    Returns:
        list: list of solutions [X, S]
    """
    # Unpack parameters
    mumax, Ks, Yxs, max_biomass = Params['mumax'], Params['Ks'], Params['Yxs'], Params['max_biomass']
    # Initial condition
    MonodVars_0 = [Initials['X0'], Initials['S0']]

    # Parameters
    args = (mumax, Ks, Yxs, max_biomass)

    # Integrate ODES
    MonodSol = spi.odeint(MonodEqnODE, MonodVars_0, t, args=args)

    # Return the result as a NumPy array
    return MonodSol

def add_NightShift(time, NightStart, NightShift):
    """
    Add a night shift to the data. The night shift is a time period where no measurements are taken.
    Args:
    time: List, time vector
    NightStart: Float, time point where the night shift starts
    NightShift: Float, time period of the night shift

    Returns:
    time: List, time vector with night shift
    """
    # if the time vector/experiment is smaller than 15h there is no night shift
    if time[-1] < 15:
        return time
    else:
        # finding all nights in the sampling period
        NightStart = np.arange(NightStart,time[-1],24)

        # Get the index of the row with the value closest to NightStart
        idx_start = np.abs(np.tile(time, (len(NightStart),1)).T - np.tile(NightStart, (len(time),1))).argmin(axis=0)
        # Get the index of the row with the value closest to NightStart+NightShift
        idx_end = np.abs(np.tile(time, (len(NightStart),1)).T - np.tile(NightStart+NightShift, (len(time),1))).argmin(axis=0)
        # storing all indices for all night shifts
        nights = np.hstack([np.arange(start,stop) for start, stop in zip(idx_start, idx_end)])
        # Delete the rows between idx_start and idx_end
        return np.delete(time, np.s_[nights])

def OD2X(OD, OD2X):
    """
    Convert OD to biomass concentration.
    """
    return OD * OD2X
def X2OD(X, OD2X):
    """
    Convert biomass concentration to OD.
    """
    return X / OD2X


#!/usr/bin/env python

"""Tests for `silvio` package."""

import pytest

import os

from silvio import first, DATADIR


def test_datadir_exists () :
    """Check if directory of DATADIR actually exists."""
    datadir_exists = os.path.isdir(DATADIR)
    assert datadir_exists == True


def test_use_an_utility () :
    """Test usage of a simple utility, just to see if import is correct."""
    el = first( ["aaa","bbb","ccc"], lambda it : it == "bbb" )
    assert el == "bbb"



"""
This is a TESTING module to showcase Host events.
The number inside this module (example: a size of a cell) will increase with the number of genes.
"""

from ...host import Host
from ...module import Module
from ...events import EventEmitter, EventLogger
from ..all_events import InsertGeneEvent, RemoveGeneEvent



class PhenotypeSize (Module) :

    size: int



    def make ( self, size:int = 0 ) -> None :
        self.size = 0



    def copy ( self, ref:'PhenotypeSize' ) -> None :
        self.size = ref.size



    def bind ( self, host:Host ) -> None :
        host.observe( InsertGeneEvent, self.listen_insert_gene )
        host.observe( RemoveGeneEvent, self.listen_remove_gene )



    def sync ( self, emit:EventEmitter, log:EventLogger ) -> None :
        pass # Nothing to sync.



    def listen_insert_gene ( self, event:InsertGeneEvent, emit:EventEmitter, log:EventLogger ) -> None :
        self.size += 1
        log("PhenotypeSize: incremented size by 1")



    def listen_remove_gene ( self, event:RemoveGeneEvent, emit:EventEmitter, log:EventLogger ) -> None :
        self.size -= 1
        log("PhenotypeSize: decremented size by 1")



"""
Sets are themed collections of various tools and utilities. They allow for developers to keep
related features in the same directory, but ideally these features should all be generalized and
distributed among the `extension.{modules|records|tools|utils}` sub-packages.

Use a `set` if you are not able to fully integrate the features in a generalist manner.
Sub-packages inside `sets` are meant to be internal only and you may make them available for public
by writing acessors in `extension.{modules|records|tools|utils}` files.
"""


# Fermentation Process Simulator
# noinspection PySingleQuotedDocstring
import os
from math import ceil
from random import uniform
import matplotlib.pyplot as plt
from datetime import datetime
import json

import numpy as np
import pandas as pd
import sklearn.metrics
from scipy.optimize import Bounds, minimize


class MonodModel:
    '''
    The 'Monod_Model' class stores all information about the bioprocess model its properties.
        Attributes:
            OperationMode: str, Mode of operation ('batch', 'fedbatch' or 'continuous')
            Params: dict, contains model parameters for monod-like kinetics, e.g. µmax
            Conditions: dict, contains starting conditions for simulated fermentation, like [Substrate]
    '''
    # Class attributes
    __Organism = 'E. coli'                          # Alternatively: 'Pput'
    __OperationMode = 'batch'                       # Alternatively: 'fedbatch', 'continuous'
    __Results = dict()

    def __init__(self):
        # Instance attributes
        self.__Params = dict(
                {
                        'u0':       0,                            # Initial growth rate [h^-1]
                        'umax':     round(uniform(0.5, 1.1), 3),  # maximal growth rate [h^-1] (default 0.5 - 1.1)
                        'duration': 24,                           # Process Duration [h] as Integer
                        'Ks':       round(uniform(7, 10), 3),
                        # Monod substrate affinity constant (default 7 - 10) [g/L]
                        'Yx':       round(uniform(0.4, 0.6), 3),
                        # Yield coefficient for growth on glucose (default 0.4 - 0.6) [g/g]
                        'k1':       round(uniform(0.05, 0.2), 3),
                        # Production rate of Product (default 0.05 - 0.2) [# h^-1]

                        # sources: https://bionumbers.hms.harvard.edu/bionumber.aspx?id=105318,
                        # https://bionumbers.hms.harvard.edu/bionumber.aspx?s=n&v=3&id=111049
                })
        self.__hiddenParams = dict(
                {
                        'u0':       0,                            # Initial growth rate [h^-1]
                        'umax':     round(uniform(0.5, 1.1), 3),  # maximal growth rate [h^-1] (default 0.5 - 1.1)
                        'duration': 24,                           # Process Duration [h] as Integer
                        'Ks':       round(uniform(7, 10), 3),
                        'Yx':       round(uniform(0.4, 0.6), 3),
                        'k1':       round(uniform(0.05, 0.2), 3),
                })
        self.__Conditions = dict(
                {
                        'S0': round(uniform(19, 21), 3),  # Initial substrate concentration [g/L]
                        'P0': 0,  # Initial product concentration [g/L]
                        'X0': round(uniform(0.05, 0.3), 3)  # Initial biomass concentration [g/L]
                })
        self.var_ModelingCount = 0
        self.var_ExpCount = 0
        self.__optimal_S = round(uniform(15, 30), 3)
        self.var_Organism = self.__Organism
        self.var_OperationMode = self.__OperationMode
        self.var_Params = self.__Params
        self.var_Conditions = self.__Conditions
        self.__TimeCreated = datetime.now().strftime('%d_%m_%Y_%I_%M_%S_%f')
        self.__ModelName = f'Monod_Model_{self.__TimeCreated}'
        self.__Description = f'Monod_Model instance for a {self.var_Organism}-{self.var_OperationMode} microbial ' \
                             f'production process. \nCurrent parameters: {self.var_Params}'

    def __str__(self):
        '''
        Use print(Monod_Model) for a quick description of the model instance.
        '''
        return self.__Description

    # Instance methods
    def get_start_params(self, hidden_params=False):
        '''
        This function calculates and returns the initial numeric derivatives of process parameters.
            Output:
                start_params: dict, contains initial process derivatives

        Test for complete set of params
        >>> TestModelBatch = MonodModel()
        >>> len(TestModelBatch.get_start_params()) == 3
        True

        Test for all fedbatch params
        >>> TestModel_Fedbatch = Monod_Model()
        >>> TestModel_Fedbatch.var_OperationMode = 'fedbatch'
        >>> len(TestModelBatch.get_start_params()) == 4
        True

        Test for correct sign of initial derivatives
        >>> init_params = TestModelBatch.get_start_params()
        >>> init_params['rX0'] >= 0 and init_params['rS0'] <= 0 and init_params['rP0'] >= 0
        True
        '''
        if hidden_params:
            u0, Yx, k1 = self.__hiddenParams['u0'], self.__hiddenParams['Yx'], self.__hiddenParams['k1']
        else:
            u0, Yx, k1 = self.var_Params['u0'], self.var_Params['Yx'], self.var_Params['k1']
        X0, S0 = self.var_Conditions['X0'], self.var_Conditions['S0']

        if self.var_OperationMode == 'batch':
            rX0 = u0 * X0                       # Initial
            rS0 = -(rX0 / (Yx / S0))            # Startwert Änderungsrate S
            rP0 = (k1 * u0) * X0                # Startwert Änderungsrate P
        else:
            rX0 = 0
            rS0 = 0
            rP0 = 0                             # TODO: Fedbatch & Konti Ergänzen!

        start_params = {
                'rX0': rX0,
                'rS0': rS0,
                'rP0': rP0
                        }

        return start_params

    def calculate_monod(self, hidden_params=False):
        '''
        Calculates monod kinetics for current model instance.
            Output:
                monod_result: dict, Result of kinetics (X, S, P, µ, rX, rS, rP) as Lists

        Test correct length of model output
        >>> TestModelBatch = Monod_Model()
        >>> monod_result = TestModelBatch.calculate_monod()
        >>> len(monod_result) == 7
        True

        >>> len(monod_result['X']) == TestModelBatch.var_Params['duration']
        True

        >>> max(monod_result['u']) <= TestModelBatch.var_Params['umax']
        True
        '''
        # Get start parameters
        params = self.get_start_params(hidden_params)
        rX, rS, rP = [params['rX0']], [params['rS0']], [params['rP0']]
        if hidden_params:
            duration, umax, Ks, = self.var_Params['duration'], self.__hiddenParams['umax'], self.__hiddenParams['Ks']
            Yx, k1, u = self.__hiddenParams['Yx'], self.__hiddenParams['k1'], [self.__hiddenParams['u0']]
        else:
            duration, umax, Ks,  = self.var_Params['duration'], self.var_Params['umax'], self.var_Params['Ks']
            Yx, k1, u = self.var_Params['Yx'], self.var_Params['k1'], [self.var_Params['u0']]
        S, P, X = [self.var_Conditions['S0']], [self.var_Conditions['P0']], [self.var_Conditions['X0']]

        for j in range(1, duration):
            new_u = umax * S[j - 1] / (Ks + S[j - 1])       # Change of µ
            if new_u >= 0:
                u.append(new_u)
            else:
                u.append(0)

            new_rX = u[j - 1] * X[j - 1]                    # Derivative of Biomass
            if new_rX >= 0:
                rX.append(new_rX)
            else:
                rX.append(0)
            X.append(X[j - 1] + rX[j])                      # New [Biomass]

            new_rS = -(rX[j - 1] / Yx)                      # Derivative of substrate
            if new_rS <= 0:
                rS.append(new_rS)
            else:
                rS.append(0)
            new_S = S[j - 1] + rS[j]
            if new_S < 0:
                S.append(0)                                 # New [Substrate]
            else:
                S.append(new_S)

            new_rP = (k1 * u[j]) * X[j]                     # Derivative of product
            if new_rP >= 0:
                rP.append(new_rP)
            else:
                rP.append(0)

            P.append(P[j - 1] + rP[j])                      # New [Product]

        monod_result = {
                'X': X,
                'S': S,
                'P': P,
                'u': u,
                'rX': rX,
                'rS': rS,
                'rP': rP
                }

        if not hidden_params:
            self.Results = monod_result

        return monod_result

    def plot_results(self, offline_results: pd.DataFrame = None):
        '''
        Returns plot of current model instance results (X,S,P vs. Time).
        Raises AttributeError if no results are stored inside the model when the method is called.

        >>> TestModelBatch = Monod_Model()
        >>> TestModelBatch.plot_results()
        Traceback (most recent call last):
        ...
        AttributeError: No Results yet! Call Monod_Model.calculate_monod() before plotting.
        '''
        if not hasattr(self, 'Results'):
            raise AttributeError('No Results yet! Call Monod_Model.calculate_monod() before plotting.')

        time = range(1, self.var_Params['duration']+1)
        X, S, P = self.Results['X'], self.Results['S'], self.Results['P']

        plt.plot(time, X, 'r', time, S, 'g', time, P, 'b')
        if offline_results is not None:
            off_time = offline_results.index + 1
            off_X, off_S, off_P = offline_results['X'].values, offline_results['S'].values, offline_results['P'].values
            plt.plot(off_time.values, off_X, 'r', off_time.values, off_S, 'g', off_time.values, off_P, 'b', linestyle='', marker='X')

        plt.legend(['Biomass [g/L]', 'Substrate [g/L]', 'Product [g/L]'])
        plt.ylabel('Biomass, Substrate & Product Concentration [g/L]')

        plt.xlabel('Process Duration [h]')
        plt.title(self.__Description)
        plt.show()
        return plt

    def calc_new_mu(self, step: int):
        '''
        Calculates new µ value depending on different inhibition terms.
        Raises AttributeError if mode of operation or inhibition term is unsupported.

        >>> TestModelBatch = Monod_Model()
        >>> TestModelBatch.var_OperationMode = 'test'
        >>> TestModelBatch.calc_new_mu(1)
        Traceback (most recent call last):
        ...
        AttributeError: Unsupported mode of operation. Check to see if Model.var_OperationMode is one of "batch", "fedbatch", or "continuous".
        '''
        self.var_Paramsparams = self.get_start_params()
        duration, umax, Ks,  = self.var_Params['duration'], self.var_Params['umax'], self.var_Params['Ks']
        # Yx, k1, u = self.var_Params['Yx'], self.var_Params['k1'], [self.var_Params['u0']]
        # TODO: Anpassen auf iteratives Aufrufen der Monod_Calculation
        S, P, X = [self.var_Conditions['S0']], [self.var_Conditions['P0']], [self.var_Conditions['X0']]

        if self.var_OperationMode == 'batch':
            new_u = umax * S[step - 1] / (Ks + S[step - 1])
        elif self.var_OperationMode == 'fedbatch':
            new_u = 0
        elif self.var_OperationMode == 'continuous':
            new_u = 0
        else:
            raise AttributeError('Unsupported mode of operation. Check to see if Model.var_OperationMode is one of '
                                 '"batch", "fedbatch", or "continuous".')

        return new_u

    def results_to_csv(self, experiments_ID: int = 0):
        '''
        Writes calculated results to .csv file in dir /csv_files
        :param experiments_ID:
        :return:
        '''
        pathname = os.path.relpath('model_results')
        if not os.path.isdir(pathname):
            os.mkdir(pathname)
        df = pd.DataFrame(self.Results)
        filename = os.path.join(pathname, f'Experiment_{experiments_ID}_{self.__ModelName}.csv')
        df.to_csv(filename)

    def to_json(self, suffix: str = None):
        '''
        Serializes model results to .json.
        '''
        pathname = os.path.relpath('json_files')
        if not os.path.isdir(pathname):
            os.mkdir(pathname)

        if suffix:
            filename = f'{self.__ModelName}_{suffix}.json'
        else:
            filename = f'{self.__ModelName}.json'
        path_file_name = os.path.join(pathname, filename)

        with open(path_file_name, 'w') as outfile:
            json.dump(self, outfile, indent=4, default=lambda o: o.__dict__)

        return path_file_name

    def from_json(self, json_path: str):
        '''
        Deserializes model from .json.
        '''
        with open(json_path, 'r') as file:
            readout = json.load(file)
            new_model = MonodModel()
            new_model.var_Params = readout['var_Params']
            new_model.var_Conditions = readout['var_Conditions']
            new_model.var_Organism = readout['var_Organism']
            new_model.var_OperationMode = readout['var_OperationMode']
            new_model.Results = readout['Results']
            new_model.__Description = readout['_MonodModel__Description']
        return new_model

    def Make_SubstrateGrowthExp(self, substrate_values: list, experiments_ID: int):
        '''

        '''
        ncols = 2
        nrows = ceil(len(substrate_values)/ncols)
        for value in substrate_values:
            assert type(value) == int or float, 'Substrate value is not a number!'
            self.var_Conditions['S0'] = value
            self.calculate_monod()
            plt = self.plot_results()
            self.to_json(suffix=str(experiments_ID))
            self.results_to_csv(experiments_ID)
            self.var_ExpCount += 1

        return plt

    def offline_samples(self, experiments_ID: int = 0):
        '''
        Simulates manual sampling of process with samples at discrete timesteps.
        :return:
        '''

        results = pd.DataFrame(self.calculate_monod(hidden_params=True))
        sampling_times = [0, 2, 4, 6, 8, 20, 23]
        offline_values = results.iloc[sampling_times][['X', 'S', 'P']]

        pathname = os.path.join('..','Data','FermProSim_samples')
        if not os.path.isdir(pathname):
            os.mkdir(pathname)

        mu, sigma = 0, 0.1
        noise = np.random.normal(mu, sigma, [offline_values.shape[0], offline_values.shape[1]])
        offline_values = offline_values + noise
        offline_values[offline_values < 0] = np.nan

        filename = os.path.join(pathname, f'Experiment_{experiments_ID}_{self.__ModelName}.csv')
        offline_values.to_csv(filename)

        time = np.array(offline_values.index)
        X, S, P = offline_values['X'].values, offline_values['S'].values, offline_values['P'].values

        plt.plot(time, X, 'r', time, S, 'g', time, P, 'b', linestyle='', marker='X')
        plt.legend(['Biomass [g/L]', 'Substrate [g/L]', 'Product [g/L]'])
        plt.ylabel('Biomass, Substrate & Product Concentration [g/L]')

        plt.xlabel('Process Duration [h]')
        plt.title(self.__Description)
        plt.show()

        return filename

    def load_offline_values(self, experiment_name):
        '''

        :param experiments_ID:
        :return:
        '''
        return pd.read_csv(experiment_name, index_col=0)

    def get_optimal_X(self):
        save_old_S = self.var_Conditions['S0']
        self.var_Conditions['S0'] = self.__optimal_S
        opt_res = self.calculate_monod(hidden_params=True)
        optimal_X = round(max(opt_res['X']), 1)
        self.var_Conditions['S0'] = save_old_S
        print(f'Desired biomass for this process is {optimal_X} g/L.')
        self.__optimal_X = optimal_X
        return optimal_X

    def calc_rmse(self, param_list=None):
        '''

        :param offline_values:
        :return:
        '''
        if param_list is not None:
            self.set_params(param_list, count=False)
        self.calculate_monod()
        calc_X, calc_S, calc_P = self.Results['X'], self.Results['S'], self.Results['P']
        right_values = self.calculate_monod(hidden_params=True)
        real_X, real_S, real_P = right_values['X'], right_values['S'], right_values['P']

        rmse = sklearn.metrics.mean_squared_error([calc_X, calc_S, calc_P], [real_X, real_S, real_P], squared=False)
        return rmse

    def fit_model(self, param_list: list):
        '''

        :param param_list:
        :param offline_values:
        :return:
        '''
        # umax, Ks, Yx, k1
        bounds = Bounds([0.5, 7, 0.4, 0.05], [1.1, 10, 0.6, 0.2])

        optimizer = minimize(self.calc_rmse, param_list, bounds=bounds)
        print(f'Model optimized in {optimizer.nit} steps.')

        return [optimizer.x[0], optimizer.x[1], optimizer.x[2], optimizer.x[3]]

    def set_params(self, param_list: list, count=True):
        '''

        :param count:
        :param param_list:
        :return:
        '''
        self.var_Params['umax'] = param_list[0]
        self.var_Params['Ks'] = param_list[1]
        self.var_Params['Yx'] = param_list[2]
        self.var_Params['k1'] = param_list[3]
        if count:
            self.var_ModelingCount += 1

    def set_conditions(self, condition_list: list, count=True):
        '''

        :param condition_list:
        :param count:
        :return:
        '''
        self.var_Conditions['S0'] = condition_list[0]
        self.var_Conditions['X0'] = condition_list[1]
        if count:
            self.var_ExpCount += 1

    def get_num_experiments(self):
        '''

        :return:
        '''

        print(f'Model parameters were changed {self.var_ModelingCount} times.\n'
              f'{self.var_ExpCount} Experiments performed.')

    def plot_linear_fit(self, offline_results: pd.DataFrame):
        '''

        :param offline_results:
        :return:
        '''
        if not hasattr(self, 'Results'):
            self.calculate_monod()

        X, S, P = self.Results['X'], self.Results['S'], self.Results['P']
        time = range(1, self.var_Params['duration'] + 1)
        plt.plot(time, X, 'r', time, S, 'g', time, P, 'b')
        off_time = offline_results.index + 1
        offline_results.fillna(0, inplace=True)
        off_X, off_S, off_P = offline_results['X'], offline_results['S'], offline_results['P']
        off_dict = {'X': [], 'S': [], 'P': []}
        for i in range(0, self.var_Params['duration']):
            if i in offline_results.index.to_list():
                off_dict['X'].append(off_X[i])
                off_dict['S'].append(off_S[i])
                off_dict['P'].append(off_P[i])
            else:
                off_dict['X'].append(np.nan)
                off_dict['S'].append(np.nan)
                off_dict['P'].append(np.nan)
        off_df = pd.DataFrame(off_dict)
        off_df.interpolate(inplace=True)
        rmse = sklearn.metrics.mean_squared_error([X, S, P], [off_df['X'], off_df['S'], off_df['P']], squared=False)

        plt.plot(off_time.values, off_X.values, 'r', off_time.values, off_S.values, 'g', off_time.values, off_P.values, 'b', linestyle='--', marker='X')

        plt.legend(['Biomass [g/L]', 'Substrate [g/L]', 'Product [g/L]'])
        plt.ylabel('Biomass, Substrate & Product Concentration [g/L]')

        plt.xlabel('Process Duration [h]')
        plt.title(f'RMSE of linear fit: {round(rmse, 3)}')
        plt.show()

        return rmse

    def get_max_biomass(self):
        if not hasattr(self, 'Results'):
            self.calculate_monod()

        X = self.Results['X']
        max_X = max(X)
        print(f'Max. biomass in current setup is: {round(max_X, 1)} g/L.')

        mae_X = sklearn.metrics.mean_absolute_error([max_X], [self.__optimal_X])
        print(f'Absolute error of current and desired biomass: {round(mae_X, 3)}')

        return max_X



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

# with open('requirements.txt') as req_file:
#     requirements = req_file.read()

with open('requirements.txt') as req_file:
    requirements = [
        line.strip() for line in req_file
        if line.strip() and not line.strip().startswith('#')
    ]

test_requirements = ['pytest>=3', ]

# getting the latest version from the __init__.py file in src/silvio
import re
import os
VERSIONFILE = os.path.join('src', 'silvio', '__init__.py')
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__=['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

setup(
    name='silvio',
    version=verstr, # update version string in src/silvio/__init__.py
    url='https://git.rwth-aachen.de/ulf.liebal/silvio.git',
    author="Ulf Liebal",
    author_email='ulf.liebal@rwth-aachen.de',
    python_requires='>=3.9', 
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Typing :: Typed',
    ],
    description="silvio is an environment for Simulation of Virtual Organisms. silvio contains several linked microbial models.",
    entry_points={},
    install_requires=requirements,
    license="MIT license",
    # long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='biotechnology, microbiology, virtual cell, systems biology',
    package_dir={'': 'src'},
    packages=find_packages(where="./src"),
    test_suite='tests',
    tests_require=test_requirements,
    zip_safe=False,
)




from typing import Dict, List, Optional, NamedTuple

import numpy as np
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from ...common import Base, ReadMethod





BaseProbDict = Dict[Base,List[float]]



class Scaffold ( NamedTuple ) :
    """
    A sequence of bases, with added information about read quality and relative distances.
    Read quality is expressed in a ratio between 0 (worst) and 1 (best).
    Can either have a single contig or 2 contigs with paired ends.
    The R2 contig has a reversed sequence.
    The expected expected_len includes both contigs and the gap.

        |  R1 contig  |            gap            |  R2 contig  |
        [0123=========] - - - - - - - - - - - - - [=========3210]
         ---> order                                   order <---
    """
    expected_len : float # Just an expectation and not necessarily integer.
    r1_seqrecord : SeqRecord
    r2_seqrecord : Optional[SeqRecord]



class LocalizedSequence ( NamedTuple ) :
    """ A sequence with a starting position. """
    sequence : Seq
    locus : int



class EstimatedSequence ( ) :
    """
    An EstimatedSequence can hold a base probability for each of the positions in the gene sequence.
    Assemblers that split their certainty of a base call between two or more different bases can
    express that with these probabilities.
    Deterministic sequences can also be converted to estimated sequences where each corresponding
    base call is rated with a full probability.
    """

    # Accessed via: base_probs[base][position] = probability-of-base-in-position
    # Values will be normalized, so it does not need to be necessarily a probability.
    base_probs : BaseProbDict



    def __init__ ( self, base_props:BaseProbDict ) :

        # Check that at least one base has been given as probabilities.
        if len(base_props.keys()) == 0 :
            raise Exception("To create an EstimatedSequence the probabilities of at least one base"
                " have to be specified.")
        first_base = next(iter(base_props.keys()))

        # Check if each sequence for a base has the same length.
        for base in base_props.keys() :
            if len(base_props[base]) != len(base_props[first_base]) :
                raise Exception("The probabilities of all bases need to have the same size.")

        self.base_probs = base_props



    def __len__ ( self ) :
        first_base = next(iter(self.base_probs.keys()))
        return len(self.base_probs[first_base])



    def calc_shannon_entropy ( self ) -> float :
        """
        Calculate the Shannon Entropy over the own estimated sequence.
        The Shannon Entropy will be calculated for each position and the average will be output.
        """
        shannon_sum = 0

        first_base = next(iter(self.base_probs.keys()))
        for i in range(len(self.base_probs[first_base])) :
            bpli = np.array([ self.base_probs[base][i] for base in self.base_probs.keys() ])
                # bpli -> base probabilities on location i
            nbpli = bpli / bpli.sum()
                # nbpli -> normalized base probabilities on location i
            with np.errstate(divide='ignore') : # Disable warning to allow log2 calculation on zero
                shannon_i = -np.sum(nbpli * np.nan_to_num( np.log2(nbpli), neginf=0.0 ) )
                    # shannon_i -> shannon entropy on location i
                    # Additional conversion from NaN to 0 is required because: 0 * log(0) == NaN
            shannon_sum += shannon_i

        shannon_avg = shannon_sum / len(self.base_probs[first_base])
        return shannon_avg



    def as_consensus_sequence ( self ) -> Seq :
        """
        Outputs the consensus sequence by choosing the base with highest probability for each location.
        """
        consensus:str = ""
        first_base = next(iter(self.base_probs.keys()))
        for i in range(len( self.base_probs[first_base] )) :
            quorum = dict([ (base,self.base_probs[base][i]) for base in self.base_probs.keys() ])
            max_base = max( quorum, key=quorum.get, default='-' ) # Get Key with maximum value. Get first if equal.
            consensus += max_base
        return Seq(consensus)



def get_consensus_from_overlap ( locseqs:List[LocalizedSequence] ) -> LocalizedSequence :
    """
    Output the consensus sequence of multiple localized sequences.
    Returns: ( min starting locus, max ending locus, consensus sequence )
    """

    # Pay attention that indexes start at zero while locus could be negative.
    min_start = min([ locseq.locus for locseq in locseqs ])
    max_end = max([ locseq.locus + len(locseq.sequence) for locseq in locseqs ])
    conseq:List[str] = [ "-" for _ in range(min_start,max_end) ]

    # For each position in the consensus sequence, get the value of each localized sequence in that
    # same position. Could be out of bounds.
    for pos in range( min_start, max_end ) :
        quorum:Dict[Base,int] = {}
        for locseq in locseqs :
            if pos >= locseq.locus and pos < locseq.locus + len(locseq.sequence) : # inside bounds

                # Add this base to the possible bases in the consensus. Initialize if required.
                cur_base = str(locseq.sequence[ pos - locseq.locus ])
                if cur_base not in quorum :
                    quorum[cur_base] = 0
                quorum[cur_base] += 1

        # Based on the quorum, select the best base for this position.
        max_base = max( quorum, key=quorum.get, default='-' ) # Get Key with maximum value. Get first if equal.
        conseq[ pos - min_start ] = max_base

    return LocalizedSequence( sequence= Seq("".join(conseq)), locus= min_start )



def estimate_from_overlap ( locseqs:List[LocalizedSequence] ) -> EstimatedSequence :
    """
    Given a list of localized sequences, overlap them and measure the probability of bases their
    sum returns.
    Similar to the `get_consensus_from_overlap` method, but the probabilities are kept instead
    of the best base.
    """

    # Pay attention that indexes start at zero while locus could be negative.
    min_start = min([ locseq.locus for locseq in locseqs ])
    max_end = max([ locseq.locus + len(locseq.sequence) for locseq in locseqs ])
    quorum:BaseProbDict = {
        'A': [ 0 for _ in range(min_start,max_end) ],
        'C': [ 0 for _ in range(min_start,max_end) ],
        'G': [ 0 for _ in range(min_start,max_end) ],
        'T': [ 0 for _ in range(min_start,max_end) ],
    }

    # For each position in the consensus sequence, get the value of each localized sequence in that
    # same position. Could be out of bounds.
    for pos in range( min_start, max_end ) :
        for locseq in locseqs :
            if pos >= locseq.locus and pos < locseq.locus + len(locseq.sequence) : # inside bounds
                cur_base = str(locseq.sequence[ pos - locseq.locus ])
                quorum[cur_base][ pos - min_start ] += 1

    # Convert the counts in the quorum into a probability. If the quorum is completely empty,
    # then the probability for each base is equalized.
    for pos in range( max_end - min_start ) :
        sum_counts = sum([ quorum[base][pos] for base in ['A','C','G','T'] ])
        if sum_counts == 0 : # Special case to set equal probabilities.
            for base in ['A','C','G','T'] :
                quorum[base][pos] = 1/4
        else : # Normal case to normalize into probabilities.
            for base in ['A','C','G','T'] :
                quorum[base][pos] = quorum[base][pos] / sum_counts

    return EstimatedSequence(quorum)







def ErrorRate(Invest, ResTotal= 5000, relKM=.005, Vmax=.9):
    '''
    Calculate the experimental error rate based on the investment to equipment. Based on inverse Michaelis-Menten equation mit max error 0.9 and min error 0.1.
    '''
    KM = relKM*ResTotal
    myMax = 1
    return myMax - (Vmax * Invest) / (KM + Invest)
